# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

NOMINAL_FOCALLENGTH = 10061.65
'''Numbers taken from the Chandra coordinate memo I

at http://cxc.harvard.edu/contrib/jcm/ncoords.ps
'''


TLMINMAX = {'ACIS': {'CCD_ID': [0, 9],
                     'CHIPX': [2, 1023],
                     'CHIPY': [2, 1023],
                     'TDETX': [2, 8191],
                     'TDETY': [2, 8191],
                     'DETX': [0.5, 8192.5],
                     'DETY': [0.5, 8192.5],
                     'X': [0.5, 8192.5],
                     'Y': [0.5, 8192.5],
                     'PHA': [0, 36855],
                     'ENERGY': [0, 1e6],  # This is the ACIS detected energy, not the real one
                     'FLTGRADE': [0, 255],
                    }
            }


AIMPOINTS = {'ACIS-I': [-.782, 0, -233.592],
             'ACIS-S': [-.684, 0, -190.133],
             'HRC-I': [-1.040, 0, 126.985],
             'HRC-S': [-1.43, 0, 250.456],
             }
'''Default aimpoints from coordiante memo.

A lot of work has gone into aimpoints since 2001, specifically because they also move on the
detector. For now, we just implement this simple look-up here, but a more detailed implementation
that reaches out to CALDB might be needed in the future.

Numbers are taken for the Chandra coordinate memo I at
http://cxc.harvard.edu/contrib/jcm/ncoords.ps
'''

TDET = {'ACIS': {'version': 'ACIS-2.2',
                 'theta': np.deg2rad(np.array([90., 270., 90., 270., 0, 0, 0, 0, 0, 0])),
                 'scale': np.ones(10),
                 'handedness': np.ones(10), # could be list of arrays
                 'origin': np.array([[3061, 5131],
                                     [5131, 4107],
                                     [3061, 4085],
                                     [5131, 3061],
                                     [ 791, 1702],
                                     [1833, 1702],
                                     [2875, 1702],
                                     [3917, 1702],
                                     [4959, 1702],
                                     [6001, 1702]])
                 },
        }
'''Constants used to transform from chip coordinates to TDET system.

Numbers are taken for the Chandra coordinate memo I at
http://cxc.harvard.edu/contrib/jcm/ncoords.ps
'''

ODET = {'ACIS': [4096.5, 4096.5]}
'''Position of coordinate origin in detector coordinates'''

PIXSIZE = {'ACIS': 0.492 / 3600.}
'''Nominal size of pixels on the sky in degrees.'''

PIX_CORNER_LSI_PAR='''
###################  LSI to CPC corners definition file ##############
#   Chip Corner LSI coords of ACIS, HRC-I, HSI, and ACIS-2C in mm    #
#                                                                    #
#   Chip  CPC XMAX   CPC YMAX                                        #
#   Corner   LSI X  LSI Y    LSI Z                                   #
#                                                                    #
# Reference: Tables 14-17, "ASC Coordinates", July 31, 1996 version  #
######################################################################

ACIS,s,h,"CORNERS AXAF-LSI-1.0/ACIS-I&S&C AXAF-CPC-1.0",,,
#====================================================================#
#
ACIS-Title,s,h,"ACIS-I/S (AI1) LSI coordinates (mm)",,,
ACIS-chip-size,s,h,"(1024, 1024)",,," ACIS chip size (x, y) in pixel"
#
ACIS-I0-LL,s,h,"( 2.361 -26.484  23.088 )",,,"LSI coords for ACIS-I0 LL"
ACIS-I0-LR,s,h,"( 1.130 -26.546  -1.458 )",,,"LSI coords for ACIS-I0 LR"
ACIS-I0-UR,s,h,"(-0.100  -2.001  -1.458 )",,,"LSI coords for ACIS-I0 UR"
ACIS-I0-UL,s,h,"( 1.130  -1.939  23.088 )",,,"LSI coords for ACIS-I0 UL"
#
ACIS-I1-LL,s,h,"( 1.130  23.086  -1.458 )",,,"LSI coords for ACIS-I1 LL"
ACIS-I1-LR,s,h,"( 2.360  23.024  23.088 )",,,"LSI coords for ACIS-I1 LR"
ACIS-I1-UR,s,h,"( 1.130  -1.521  23.088 )",,,"LSI coords for ACIS-I1 UR"
ACIS-I1-UL,s,h,"(-0.100  -1.459  -1.458 )",,,"LSI coords for ACIS-I1 UL"
#
ACIS-I2-LL,s,h,"( 1.130 -26.546  -1.997 )",,,"LSI coords for ACIS-I2 LL"
ACIS-I2-LR,s,h,"( 2.361 -26.484 -26.543 )",,,"LSI coords for ACIS-I2 LR"
ACIS-I2-UR,s,h,"( 1.130  -1.939 -26.543 )",,,"LSI coords for ACIS-I2 UR"
ACIS-I2-UL,s,h,"(-0.10   -2.001  -1.997 )",,,"LSI coords for ACIS-I2 UL"
#
ACIS-I3-LL,s,h,"( 2.361  23.024 -26.543 )",,,"LSI coords for ACIS-I3 LL"
ACIS-I3-LR,s,h,"( 1.131  23.086  -1.997 )",,,"LSI coords for ACIS-I3 LR"
ACIS-I3-UR,s,h,"(-0.100  -1.459  -1.997 )",,,"LSI coords for ACIS-I3 UR"
ACIS-I3-UL,s,h,"( 1.13   -1.521 -26.543 )",,,"LSI coords for ACIS-I3 UL"
#
ACIS-S0-LL,s,h,"( 0.744 -81.170 -59.170 )",,,"LSI coords for ACIS-S0 LL"
ACIS-S0-LR,s,h,"( 0.353 -56.597 -59.170 )",,,"LSI coords for ACIS-S0 LR"
ACIS-S0-UR,s,h,"( 0.353 -56.597 -34.590 )",,,"LSI coords for ACIS-S0 UR"
ACIS-S0-UL,s,h,"( 0.744 -81.170 -34.590 )",,,"LSI coords for ACIS-S0 UL"
#
ACIS-S1-LL,s,h,"(0.348  -56.133 -59.170 )",,,"LSI coords for ACIS-S1 LL"
ACIS-S1-LR,s,h,"(0.099  -31.559 -59.170 )",,,"LSI coords for ACIS-S1 LR"
ACIS-S1-UR,s,h,"(0.099  -31.559 -34.590 )",,,"LSI coords for ACIS-S1 UR"
ACIS-S1-UL,s,h,"(0.348  -56.133 -34.590 )",,,"LSI coords for ACIS-S1 UL"
#
ACIS-S2-LL,s,h,"( 0.096 -31.100 -59.170 )",,,"LSI coords for ACIS-S2 LL"
ACIS-S2-LR,s,h,"(-0.011  -6.524 -59.170 )",,,"LSI coords for ACIS-S2 LR"
ACIS-S2-UR,s,h,"(-0.011  -6.524 -34.590 )",,,"LSI coords for ACIS-S2 UR"
ACIS-S2-UL,s,h,"( 0.096 -31.100 -34.590 )",,,"LSI coords for ACIS-S2 UL"
#
ACIS-S3-LL,s,h,"(-0.011  -6.035 -59.170 )",,,"LSI coords for ACIS-S3 LL"
ACIS-S3-LR,s,h,"( 0.024  18.541 -59.170 )",,,"LSI coords for ACIS-S3 LR"
ACIS-S3-UR,s,h,"( 0.024  18.541 -34.590 )",,,"LSI coords for ACIS-S3 UR"
ACIS-S3-UL,s,h,"(-0.011  -6.035 -34.590 )",,,"LSI coords for ACIS-S3 UL"
#
ACIS-S4-LL,s,h,"( 0.026  18.970 -59.170 )",,,"LSI coords for ACIS-S4 LL"
ACIS-S4-LR,s,h,"( 0.208  43.545 -59.170 )",,,"LSI coords for ACIS-S4 LR"
ACIS-S4-UR,s,h,"( 0.208  43.545 -34.590 )",,,"LSI coords for ACIS-S4 UR"
ACIS-S4-UL,s,h,"( 0.026  18.970 -34.590 )",,,"LSI coords for ACIS-S4 UL"
#
ACIS-S5-LL,s,h,"( 0.208  43.986 -59.170 )",,,"LSI coords for ACIS-S5 LL"
ACIS-S5-LR,s,h,"( 0.528  68.560 -59.170 )",,,"LSI coords for ACIS-S5 LR"
ACIS-S5-UR,s,h,"( 0.528  68.560 -34.590 )",,,"LSI coords for ACIS-S5 UR"
ACIS-S5-UL,s,h,"( 0.208  43.986 -34.590 )",,,"LSI coords for ACIS-S5 UL"
#
#
HRC,s,h,"CORNERS AXAF-LSI-1.0/HRC-I&S AXAF-CPC-1.0",,,
#
#====================================================================#
#
#Updated on 12/24/97
HRC-Title,s,h,"HRC LSI coordinates (mm)",,,
HRC-I-chip-size,s,h,"(16384, 16384)",,,"HRC-I chip size (x, y) in pixel"
HRC-S-chip-size,s,h," ( 4096, 16456)",,," HRC-S1 & S3 chip size (x, y) in pixel"
HRC-S2-chip-size,s,h,"( 4096, 16456)",,,"HRC-S2 chip size (x, y) in pixel"
#
HRC-I-LL,s,h,"(0.0      0.000   74.482 )",,,"LSI coords for HRC-I LL"
HRC-I-LR,s,h,"(0.0     74.482    0.000 )",,,"LSI coords for HRC-I LR"
HRC-I-UR,s,h,"(0.0      0.000  -74.482 )",,,"LSI coords for HRC-I UR"
HRC-I-UL,s,h,"(0.0    -74.482    0.000 )",,,"LSI coords for HRC-I UL"
#
HRC-S1-LL,s,h,"(2.644  161.949  -13.167 )",,,"LSI coords for HRC-S1 LL"
HRC-S1-LR,s,h,"(2.644  161.949   13.167 )",,,"LSI coords for HRC-S1 LR"
HRC-S1-UR,s,h,"(0.000   56.180   13.167 )",,,"LSI coords for HRC-S1 UR"
HRC-S1-UL,s,h,"(0.000   56.180  -13.167 )",,,"LSI coords for HRC-S1 UL"
#
HRC-S2-LL,s,h,"(0.000   56.180  -13.167 )",,,"LSI coords for HRC-S3 LL"
HRC-S2-LR,s,h,"(0.000   56.180   13.167 )",,,"LSI coords for HRC-S3 LR"
HRC-S2-UR,s,h,"(0.000  -49.622   13.167 )",,,"LSI coords for HRC-S3 UR"
HRC-S2-UL,s,h,"(0.000  -49.622  -13.167 )",,,"LSI coords for HRC-S3 UL"
#
HRC-S3-LL,s,h,"(0.000  -49.622  -13.167 )",,,"LSI coords for HRC-S2 LL"
HRC-S3-LR,s,h,"(0.000  -49.622   13.167 )",,,"LSI coords for HRC-S2 LR"
HRC-S3-UR,s,h,"(2.253 -155.400   13.167 )",,,"LSI coords for HRC-S2 UR"
HRC-S3-UL,s,h,"(2.253 -155.400  -13.167 )",,,"LSI coords for HRC-S2 UL"
'''

def chip2tdet(chip, tdet, id_num):
    '''Convert CHIP coordinates to TDET coordiantes.

    See eqn (5) in `the Chandra coordiante memo <http://cxc.harvard.edu/contrib/jcm/ncoords.ps>`_.

    Parameters
    ----------
    chip : (N, 2) np.array
        chip coordiantes
    tdet : dict
        dictionary with definitions for the coordiante conversion.
        See `ACISTDET` for an example.
    id_num : integer
        chip ID number (e.g. ``1`` for ACIS-I1)
    '''
    scale = tdet['scale'][id_num]
    handedness = tdet['handedness'][id_num]
    origin_tdet = tdet['origin'][id_num]
    theta = tdet['theta'][id_num]
    rotation = np.array([[np.cos(theta), np.sin(theta)],
                         [-np.sin(theta), np.cos(theta)]])
    return scale * handedness * np.dot(rotation, (chip - 0.5).T).T + (origin_tdet + 0.5)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from .base import FlatOpticalElement
from ..math.utils import h2e
from ..visualization.utils import plane_with_hole


class Baffle(FlatOpticalElement):
    '''Plate with rectangular hole that allows photons through.

    The probability of photons that miss is set to 0.

    Parameters
    ----------
    photons: astropy Table
        table that includes information on all of the photons
    '''

    display = {'color': (1., 0.5, 0.4),
               'outer_factor': 3,
               'shape': 'plane with hole'}


    def process_photons(self, photons, intersect, intercoos, interpoos):
        photons['pos'][intersect] = intercoos[intersect]
        photons['probability'][~intersect] = 0
        return photons

    def triangulate_inner_outer(self):
        '''Return a triangulation of the baffle hole embedded in a square.

        The size of the outer square is determined by the ``'outer_factor'`` element
        in ``self.display``.

        Returns
        -------
        xyz : np.array
            Numpy array of vertex positions in Eukeldian space
        triangles : np.array
            Array of index numbers that define triangles
        '''
        r_out = self.display.get('outer_factor', 3)
        g = self.geometry
        outer = h2e(g['center']) + r_out * np.vstack([h2e( g['v_x']) + h2e(g['v_y']),
                                                      h2e(-g['v_x']) + h2e(g['v_y']),
                                                      h2e(-g['v_x']) - h2e(g['v_y']),
                                                      h2e( g['v_x']) - h2e(g['v_y'])
        ])
        inner =  h2e(g['center']) + np.vstack([h2e( g['v_x']) + h2e(g['v_y']),
                                               h2e(-g['v_x']) + h2e(g['v_y']),
                                               h2e(-g['v_x']) - h2e(g['v_y']),
                                               h2e( g['v_x']) - h2e(g['v_y'])
        ])
        return plane_with_hole(outer, inner)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from transforms3d.affines import decompose
from transforms3d.euler import euler2mat

from ...utils import generate_test_photons
from .. import Parallel, ParallelCalculated
from ...optics import FlatDetector as CCD
from ...design import RowlandTorus, RowlandCircleArray

def test_Parallel_numbering():
    '''Test automatic numbering and numbers assigned to photons.
    '''
    pos = [[0, -10.1, -10.1],[0, .1, -10.1],[0, -10.1, .1],[0, .1, .1]]
    detect = Parallel(elem_class=CCD, elem_args={'pixsize': 0.01, 'zoom': 5},
                      elem_pos={'position': pos}, id_col='CCD_ID')

    for i in range(len(pos)):
        assert detect.elements[i].id_num == i

    photons = generate_test_photons(5)
    photons = detect(photons)
    assert np.all(photons['CCD_ID'] == 3.)

def test_Parallel_moving():
    '''Test moving center of structure after generation.
    '''
    pos = [[0, -10.1, -10.1],[0, .1, -10.1],[0, -10.1, .1],[0, .1, .1]]
    detkwargs = {'elem_class': CCD,
                 'elem_args': {'pixsize': 0.01, 'zoom': 5},
                 'elem_pos': {'position': pos},
                 'id_col': 'CCD_ID',
                 }
    detect = Parallel(**detkwargs)
    detect1 = Parallel(**detkwargs)

    move = np.eye(4)
    move[2, 3] = 15

    detect1.move_center(move)

    # Check that the default positions are the same
    for i in range(len(detect.elements)):
        assert np.allclose(detect.elements[i].pos4d,
                           detect1.elements[i].pos4d)

    # but rotations are applied around a different point
    rot = np.eye(4)
    rot[:3, :3] = euler2mat(.3, .2, 1.2)
    detect.uncertainty = rot
    detect.generate_elements()

    detect1.uncertainty = rot
    detect1.generate_elements()

    for i in range(len(detect.elements)):
        assert ~np.allclose(detect.elements[i].pos4d,
                            detect1.elements[i].pos4d)

def test_Parallel_offset_numbering():
    '''The randomly generated positions might generate overlapping elements,
    but that is OK to test the numbering scheme.'''
    pos = np.random.rand(5, 3) * 50
    detect = Parallel(elem_class=CCD, elem_args={'pixsize': 0.01, 'zoom': 0.1},
                      elem_pos={'position': pos}, id_num_offset=100)

    for i in range(pos.shape[0]):
        assert detect.elements[i].id_num == i + 100

def test_parallel_calculated_normals():
    '''Regression test: At one point the rotation matrix was reversed,
    mixing active and passive rotations. As a result, the detector elements
    did not match up on the Rowland circle.'''
    rowland = RowlandTorus(5900., 6020.)
    detccdargs = {'pixsize': 0.024,'zoom': [1, 24.576, 12.288]}
    det = RowlandCircleArray(rowland=rowland, elem_class=CCD,
                             elem_args=detccdargs,
                             d_element=49.652, theta=[np.pi - 0.2, np.pi -0.19])
    e1 = det.elements[0]
    e2 = det.elements[1]
    # These two edges should be close together:
    edge1 = e1.geometry['center'] + e1.geometry['v_y']
    edge2 = e2.geometry['center'] - e2.geometry['v_y']
    assert np.all(np.abs(edge1 - edge2) < 3.)

def test_parallel_calculated_rotations():
    '''Regression test #164: An index mix-up in calculate_elempos introduced
    unintended zoom and shear in the elements.
    '''
    t = np.array([[0,0,0, 1]])
    obj = ParallelCalculated(pos_spec=t,
                             normal_spec=np.array([0, 0, 1, 1]),
                             parallel_spec=np.array([.3, .4, .5, 1]),
                             elem_class=CCD)
    trans, rot, zoom, shear = decompose(obj.elem_pos[0])
    assert np.allclose(t[0, :3], trans)
    assert np.allclose(zoom, 1.)
    assert np.allclose(shear, 0.)



# Licensed under GPL version 3 - see LICENSE.rst
'''
Sources generate photons with all photon properties: energy, direction, time, polarization.

For objects of type `AstroSource`, the coordinates of the photon origin on the sky are added to the photon list. `astropy.coords.SkyCoord` is an object well suited for this task. These objects can be added to photon tables through the mechanims of `mixin columns <http://docs.astropy.org/en/latest/table/index.html#mixin-columns>`). However, mix-in columns don't (yet) support saving to all table formats or tables operations such as stacking. Thus, it is much better to include the coordinates as two columns of floats with names ``ra`` and ``dec`` into the table.

Sources take a `~astropy.coordinates.SkyCoord` from the user to avoid any ambiguity about the coordinate systme used, but convert this into plain floats used in the photon table.
'''

import os
from datetime import datetime

import numpy as np
from scipy.stats import expon
from astropy.table import Table
import astropy.units as u
from astropy.coordinates import SkyCoord, SkyOffsetFrame

from ..base import SimulationSequenceElement
from ..math.random import RandomArbitraryPdf
from .. import __version__ as marxsversion


@u.quantity_input
def poisson_process(rate: (u.s * u.cm**2)**(-1)):
    '''Return a function that generates Poisson distributed times with rate ``rate``.

    Parameters
    ----------
    rate :  `~astropy.units.quantity.Quantity`
        Expectation value for the total rate of photons with unit 1 / cm**2 / s.

    Returns
    -------
    poisson_rate : function
        Function that generates Poisson distributed times with rate ``rate``.
    '''
    if not rate.isscalar:
        raise ValueError('"rate" must be scalar.')

    @u.quantity_input(exposuretime=u.s)
    def poisson_rate(exposuretime: u.s, geomarea: u.cm**2) -> u.s:
        '''Generate Poisson distributed times.

        Parameters
        ----------
        exposuretime : `~astropy.units.quantity.Quantity`
            Exposure time
        geomarea : `~astropy.units.quantity.Quantity`
            Geometric opening area of telescope

        Returns
        -------
        times :  `~astropy.units.quantity.Quantity`
            Poisson distributed times.
        '''
        fullrate = rate * geomarea
        # Make 10 % more numbers then we expect to need, because it's random
        times = expon.rvs(scale=1./fullrate.to(1 / u.s),
                          size=int((exposuretime * fullrate * 1.1).to(u.dimensionless_unscaled)))
        # If we don't have enough numbers right now, add some more.
        while (times.sum() * u.s) < exposuretime:
            times = np.hstack([times, expon.rvs(scale=1/fullrate.to(1 / u.s),
                                                size=int(((exposuretime - times.sum() * u.s) * fullrate * 1.1).to(u.dimensionless_unscaled)))])
        times = np.cumsum(times) * u.s
        return times[times < exposuretime]
    return poisson_rate


class SourceSpecificationError(Exception):
    pass


class Source(SimulationSequenceElement):
    '''Base class for all photons sources.

    This class provides a very general implementation of photons
    sources. Typically, it is not used directly, but a more specialized
    subclass, such as `PointSource` for an astronomical source or
    `LabPointSource` for a source at a finite distance.

    Most of the derived source support the same input argumets as `Source`,
    thus they are explained in detail here.

    Parameters
    ----------
    flux : `~astropy.units.quantity.Quantity` or callable
        This sets the total flux from a source in
        photons/time/area.  Options are:

        - quantity: Constant (not Poisson distributed) flux.
        - callable: Function that takes a total exposure time as input and
          returns an array
          of photon emission times between 0 and the total exposure time.

    energy : `~astropy.units.quantity.Quantity` or callable or `~astropy.table.QTable`

        This input decides the energy of the emitted photons.
        Possible formats are:

        - polarization.
        - `~astropy.units.quantity.Quantity`: Constant energy.
        - `astropy.table.Table`:
          Given this table, the code assumes a piecewise flat spectrum.
          The "energy" values contain the **upper** limit of each bin,
          the "fluxdensity" array the flux density in each bin.
          The first entry in the "fluxdensity" array is ignored, because the lower
          bound of this bin is undefined.
          The code draws an energy from this spectrum for every photon created.
        - A function or callable object: This option allows for full
          customization. The function must take an array of photon times as
          input and return an equal length array of photon energies
          `~astropy.units.quantity.Quantity`.

    polarization : `~astropy.units.quantity.Quantity`, ``None``, `~astropy.table.QTable`, or callable.
        There are several different ways to set the polarization angle of the
        photons for a polarized source. In all cases, the angle is measured
        North through East. (We ignore the special case of a polarized source
        exactly on a pole.)
        The default value is ``None`` (unpolarized source).

        - ``None`` : An unpolarized source. Every photons is assigned a random
          polarization.
        - `~astropy.units.quantity.Quantity` :
          Constant polarization angle for all photons
        - `~astropy.table.Table` :
          Table with columns called "angle" and "probabilitydensity".
          The summed probability density will automatically be
          normalized to one. Given this table, the
          code assumes a piecewise constant probability density. The "angle"
          values contain the **upper** limit of each bin. The first entry in the
          "probabilitydenisty" array is ignored, because the lower bound of this bin
          is undefined.
        - a callable (function or callable object): This
          option allows full customization.  The function is called with two
          arrays (time and energy values) as input and must return an array of
          equal length that contains the polarization angles as
          `~astropy.units.quantity.Quantity` object.

    geomarea : `astropy.units.Quantity` or ``None``
          Geometric opening area of telescope. If ``None`` then the flux must
          be given in photons per time, not per time per unit area.

    '''
    def __init__(self, energy=1*u.keV, flux=1 / u.s / u.cm**2,
                 polarization=None, geomarea=1*u.cm**2, **kwargs):
        self.energy = energy
        self.flux = flux
        self.polarization = polarization
        self.geomarea = 1 if geomarea is None else geomarea

        super(Source, self).__init__(**kwargs)

    def __call__(self, *args, **kwargs):
        return self.generate_photons(*args, **kwargs)

    @u.quantity_input()
    def generate_times(self, exposuretime: u.s):
        if callable(self.flux):
            return self.flux(exposuretime, self.geomarea)
        elif hasattr(self.flux, 'isscalar') and self.flux.isscalar:
            return np.arange(0, exposuretime.to(u.s).value,
                             1. / (self.flux * self.geomarea * u.s).decompose()) * u.s
        else:
            raise SourceSpecificationError('`flux` must be a quantity or a callable.')

    @u.quantity_input()
    def generate_energies(self, t: u.s) -> u.keV:
        n = len(t)
        # function
        if callable(self.energy):
            en = self.energy(t)
            if len(en) != n:
                raise SourceSpecificationError('`energy` has to return an array of same size as input time array.')
            else:
                return en
        # astropy.table.QTable
        elif hasattr(self.energy, 'columns'):
            x = self.energy['energy'].to(u.keV).value
            y = (self.energy['fluxdensity'][1:]).to((u.s * u.cm**2 * u.keV)**(-1)).value
            y = np.hstack(([0], y))
            rand = RandomArbitraryPdf(x, y)
            return rand(n) * u.keV
        # scalar quantity
        elif hasattr(self.energy, 'isscalar') and self.energy.isscalar:
            return np.ones(n) * self.energy.to(u.keV,
                                               equivalencies=u.spectral())
        # anything else
        else:
            raise SourceSpecificationError('`energy` must be Quantity, function, or have columns "energy" and "fluxdensity".')

    @u.quantity_input()
    def generate_polarization(self, times: u.s, energies: u.keV) -> u.rad:
        n = len(times)
        # function
        if callable(self.polarization):
            pol = self.polarization(times, energies)
            if len(pol) != n:
                raise SourceSpecificationError('`polarization` has to return an array of same size as input time and energy arrays.')
            else:
                return pol
        elif self.polarization is None:
            return np.random.uniform(0, 2 * np.pi, n) * u.rad
        # astropy.table.QTable
        elif hasattr(self.polarization, 'columns'):
            x = self.polarization['angle'].to(u.rad).value
            y = (self.polarization['probabilitydensity']).to(1/u.rad).value
            rand = RandomArbitraryPdf(x, y)
            return rand(n) * u.rad
        # scalar quantity
        elif hasattr(self.polarization, 'isscalar') and self.polarization.isscalar:
            return np.ones(n) * self.polarization
        else:
            raise SourceSpecificationError('`polarization` must be number (angle), callable, None (unpolarized), 2.n array or have fields "angle" (in rad) and "probability".')

    def generate_photon(self):
        raise NotImplementedError

    @u.quantity_input()
    def generate_photons(self, exposuretime: u.s):
        '''Central function to generate photons.

        Calling this function generates a photon table according to
        the `flux`, `energy`, and `polarization` of this source. The
        number of photons depends on the total exposure time, which is
        a parameter of this function. Depending on the setting for
        `flux` the photons could be distributed equally over the
        interval 0..exposuretime or follow some other distribution.

        Parameters
        ----------
        exposuretime : `astropy.quantity.Quantity`
            Total exposure time.

        Returns
        -------
        photons : `astropy.table.Table`
            Table with photon properties.

        '''
        times = self.generate_times(exposuretime)
        energies = self.generate_energies(times)
        pol = self.generate_polarization(times, energies)
        n = len(times)
        photons = Table([times.to(u.s).value,
                         energies.to(u.keV).value,
                         pol.to(u.rad).value,
                         np.ones(n)],
                        names=['time', 'energy', 'polangle', 'probability'])
        photons.meta['EXTNAME'] = 'EVENTS'
        photons.meta['EXPOSURE'] = (exposuretime.to(u.s),
                                    'total exposure time [s]')

        #photons.meta['DATE-OBS'] =
        photons.meta['CREATOR'] = 'MARXS - Version {0}'.format(marxsversion)
        photons.meta["LONGSTRN"] = ("OGIP 1.0", "The OGIP long string convention may be used.")
        photons.meta['MARXSVER'] = (marxsversion, 'MARXS version')
        now = datetime.now()
        photons.meta['SIMDATE'] = (str(now.date()), 'Date simulation was run')
        photons.meta['SIMTIME'] = (str(now.time())[:10], 'Time simulation was started')
        photons.meta['SIMUSER'] = (os.environ.get('USER', 'unknown user'),
                                   'User running simulation')
        photons.meta['SIMHOST'] = (os.environ.get('HOST', 'unknown host'),
                                   'Host system running simulation')
        photons['time'].unit = u.s
        photons['energy'].unit = u.keV
        photons['polangle'].unit = u.rad
        return photons


class AstroSource(Source):
    '''Astrophysical source with a sky position

    Parameters
    ----------
    coords : `astropy.coordinates.SkySoord` (preferred)
        Position of the source on the sky. If ``coords`` is not a
        `~astropy.coordinates.SkyCoord` object itself, it is used to
        initialize such an object. See `~astropy.coordinates.SkyCoord`
        for a description of allowed input values.
    '''
    def __init__(self, **kwargs):
        coords = kwargs.pop('coords')
        if isinstance(coords, SkyCoord):
            self.coords = coords
        else:
            self.coords = SkyCoord(coords)

        if not self.coords.isscalar:
            raise ValueError("Coordinate must be scalar, not array.")
        super(AstroSource, self).__init__(**kwargs)

    def set_pos(self, photons, coo):
        '''Set Ra, Dec of photons in table

        This function write Ra, Dec to a table. It is defined here to
        make the way `astropy.coordinates.SkyCoord` objects are stored
        more uniform.  Currently, mixin columns in tables have some
        disadvantages, e.g. they cause errors on writing and on
        stacking. Thus, we store the coordinates as plain
        numbers. Since that format is not unique (e.g. units could be
        deg or rad), system could be ICRS, FK4, FK5 or other this
        conversion is done here for all astrononimcal sources.  This
        also makes it easier to change that design in the future.

        Parameters
        ----------
        photons : `astropy.table.Table`
            Photon table. Columns ``ra`` and ``dec`` will be added or overwritten.
        coo : `astropy.coords.SkyCoord`
            Photon coordinates

        '''
        photons['ra'] = coo.icrs.ra.deg
        photons['dec'] = coo.icrs.dec.deg
        photons['ra'].unit = u.degree
        photons['dec'].unit = u.degree
        photons.meta['COORDSYS'] = ('ICRS', 'Type of coordinate system')


class PointSource(AstroSource):
    '''Astrophysical point source.

    Parameters
    ----------
    kwargs : see `Source`
        Other keyword arguments include ``flux``, ``energy`` and ``polarization``.
        See `Source` for details.
    '''
    def __init__(self, **kwargs):
        super(PointSource, self).__init__(**kwargs)

    @u.quantity_input
    def generate_photons(self, exposuretime: u.s):
        photons = super(PointSource, self).generate_photons(exposuretime)
        self.set_pos(photons, self.coords)
        return photons


class RadialDistributionSource(AstroSource):
    ''' Base class for sources where photons follow some radial distribution on the sky

    Parameters
    ----------
    radial_distribution : callable
        A function that takes an interger as input, which specifies the number
        of photons to produce. The output must be an `astropy.units.Quantity`` object
        with n angles in it.
    func_par : object
        ``radial_distribution`` has access to ``self.func_par`` to hold function
        parameters. This could be, e.g. a tuple with coeffications.
    kwargs : see `Source`
        Other keyword arguments include ``flux``, ``energy`` and ``polarization``.
        See `Source` for details.
    '''
    def __init__(self, **kwargs):
        self.func = kwargs.pop('radial_distribution')
        self.func_par = kwargs.pop('func_par', None)
        super(RadialDistributionSource, self).__init__(**kwargs)

    @u.quantity_input
    def generate_photons(self, exposuretime: u.s):
        '''Photon positions are generated in a frame that is centered on the
        coordinates set in ``coords``, then they get transformed into the global sky
        system.
        '''
        photons = super(RadialDistributionSource, self).generate_photons(exposuretime)

        relative_frame = SkyOffsetFrame(origin=self.coords)
        n = len(photons)
        phi = np.random.rand(n) * 2. * np.pi * u.rad
        d = self.func(n)
        relative_coords = SkyCoord(d * np.sin(phi), d * np.cos(phi), frame=relative_frame)
        origin_coord = relative_coords.transform_to(self.coords)
        self.set_pos(photons, origin_coord)

        return photons


class SphericalDiskSource(RadialDistributionSource):
    '''Astrophysical source with the shape of a circle or ring.

    The `DiskSource` makes a small angle approximation. In contrast, this source
    implements the full spherical geometry at the cost of running slower.
    For radii less than a few degrees the difference is negligible and we recommend
    use of the faster `DiskSource`.

    Parameters
    ----------
    a_inner, a_outer : `astropy.coordinates.Angle`
        Inner and outer angle of the ring (e.g. in arcsec).
        The default is a disk with no inner hole (``a_inner`` is set to zero.)
    '''
    def __init__(self, **kwargs):
        kwargs['func_par'] = [kwargs.pop('a_outer'),
                              kwargs.pop('a_inner', 0. * u.rad)]
        kwargs['radial_distribution'] = self.radial_distribution
        super(SphericalDiskSource, self).__init__(**kwargs)

    def radial_distribution(self, n):
        '''Radial distribution function.

        See http://6degreesoffreedom.co/circle-random-sampling/ for an explanation
        of how to derive the formula. Note however, that here we use sin instead of cos
        because we measure the angle from the top.
        '''
        u = np.random.rand(n)
        return np.arccos(np.cos(self.func_par[1]) * (1. - u) + u * np.cos(self.func_par[0]))


class DiskSource(RadialDistributionSource):
    '''Astrophysical source with the shape of a circle or ring.

    This source uses a small angle approximation which is valid for radii less than
    a few degrees and runs much faster. See ``SphericalDiskSource`` for an
    implementation using full spherical geometry.

    Parameters
    ----------
    a_inner, a_outer : `astropy.coordinates.Angle`
        Inner and outer angle of the ring (e.g. in arcsec).
        The default is a disk with no inner hole (``a_inner`` is set to zero.)
    '''
    def __init__(self, **kwargs):
        kwargs['func_par'] = [kwargs.pop('a_outer'),
                              kwargs.pop('a_inner', 0. * u.rad)]
        kwargs['radial_distribution'] = lambda n: np.sqrt(self.func_par[1]**2 +
                          np.random.rand(n) * (self.func_par[0]**2 - self.func_par[1]**2))
        super(DiskSource, self).__init__(**kwargs)


class GaussSource(AstroSource):
    '''Astrophysical source with a Gaussian brightness profile.

    This source uses a small angle approximation which is valid for radii less than
    a few degrees.

    Parameters
    ----------
    sigma : `astropy.coordinates.Angle`
        Gaussian sigma setting the width of the distribution
    '''
    def __init__(self, **kwargs):
        self.sigma = kwargs.pop('sigma')
        super(GaussSource, self).__init__(**kwargs)

    @u.quantity_input()
    def generate_photons(self, exposuretime: u.s):
        '''Photon positions are generated in a frame that is centered on the
        coordinates set in ``coords``, then they get transformed into the global sky
        system.
        '''
        photons = super(GaussSource, self).generate_photons(exposuretime)

        relative_frame = SkyOffsetFrame(origin=self.coords)
        n = len(photons)
        relative_coords = SkyCoord(np.random.normal(scale=self.sigma.value, size=n) * self.sigma.unit,
                                   np.random.normal(scale=self.sigma.value, size=n) * self.sigma.unit,
                                   frame=relative_frame)
        origin_coord = relative_coords.transform_to(self.coords)
        self.set_pos(photons, origin_coord)

        return photons


class SymbolFSource(AstroSource):
    '''Source shaped like the letter F.

    This source provides a non-symmetric source for testing purposes.

    Parameters
    ----------
    size : `astropy.units.quantity`
        angular size
    kwargs : see `Source`
        Other keyword arguments include ``flux``, ``energy`` and ``polarization``.
        See `Source` for details.
    '''
    def __init__(self, **kwargs):
        self.size = kwargs.pop('size', 1. * u.degree)
        super(SymbolFSource, self).__init__(**kwargs)

    @u.quantity_input()
    def generate_photons(self, exposuretime: u.s):
        photons = super(SymbolFSource, self).generate_photons(exposuretime)
        n = len(photons)
        elem = np.random.choice(3, size=n)

        ra = np.ones(n) * self.coords.icrs.ra
        dec = np.ones(n) * self.coords.icrs.dec
        size = self.size
        ra[elem == 0] += size * np.random.random(np.sum(elem == 0))
        ra[elem == 1] += size
        dec[elem == 1] += 0.5 * size * np.random.random(np.sum(elem == 1))
        ra[elem == 2] += 0.8 * size
        dec[elem == 2] += 0.3 * size * np.random.random(np.sum(elem == 2))

        self.set_pos(photons, SkyCoord(ra, dec, frame=self.coords))

        return photons



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

def closeornan(a, b):
    '''Compare two arrays and allow nan==nan'''
    isnan = np.isnan(a)
    isnanb = np.isnan(b)
    if np.all(isnan == isnanb):
        return np.allclose(a[~isnan], b[~isnan])
    else:
        return False



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import pytest

from ...utils import generate_test_photons
from ..filter import EnergyFilter, GlobalEnergyFilter

@pytest.mark.parametrize("filterclass", [EnergyFilter, GlobalEnergyFilter])
def test_energydependentfilter(filterclass):
    '''Check energy dependent filter'''
    # Useless filter function, but good for testing
    f = filterclass(filterfunc=lambda x: 1./x)
    photons = generate_test_photons(5)
    photons['probability'] = 0.8
    photons['energy'] = np.arange(1., 6.)
    photons = f(photons)
    assert np.allclose(photons['probability'] , 0.8 / np.arange(1., 6.))

@pytest.mark.parametrize("filterclass", [EnergyFilter, GlobalEnergyFilter])
def test_energyfilter_error(filterclass):
    '''Check that Error is raised if probability is > 1 or negative.'''
    photons = generate_test_photons(1)
    for prob in [-0.1, 1.5]:
        f = filterclass(filterfunc=lambda x: np.ones_like(x) * prob)
        with pytest.raises(ValueError) as e:
            temp = f(photons)
        assert 'Probabilities returned by filterfunc' in str(e.value)

def test_difference_between_global_and_flat_filters():
    '''The global filter does not have a position in space, it just filters all
    photons coming its way. In contrast, photons can pass by the normal filter.'''
    photons = generate_test_photons(5)
    photons['pos'][:, 1] = np.arange(5)

    globalfilter = GlobalEnergyFilter(filterfunc=lambda x: 0.5)
    localfilter = EnergyFilter(filterfunc=lambda x: 0.5, zoom=2.1)

    pg = globalfilter(photons.copy())
    assert np.allclose(pg['probability'] , 0.5)

    pl = localfilter(photons.copy())
    assert set(pl['probability']) == set([0.5, 1.0])



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else.
# See astropy.sphinx.conf for which values are set there.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('..'))

# IMPORTANT: the above commented section was generated by sphinx-quickstart, but
# is *NOT* appropriate for astropy or Astropy affiliated packages. It is left
# commented out with this explanation to make it clear why this should not be
# done. If the sys.path entry above is added, when the astropy.sphinx.conf
# import occurs, it will import the *source* version of astropy instead of the
# version installed (if invoked as "make html" or directly with sphinx), or the
# version in the build directory (if "python setup.py build_sphinx" is used).
# Thus, any C-extensions that are needed to build the documentation will *not*
# be accessible, and the documentation will not build correctly.

import os
import sys
import datetime
from importlib import import_module

try:
    from sphinx_astropy.conf.v1 import *  # noqa
except ImportError:
    print('ERROR: the documentation requires the sphinx-astropy package to be installed')
    sys.exit(1)

# Get configuration information from setup.cfg
from configparser import ConfigParser
conf = ConfigParser()

conf.read([os.path.join(os.path.dirname(__file__), '..', 'setup.cfg')])
setup_cfg = dict(conf.items('metadata'))


# -- General configuration ----------------------------------------------------

# By default, highlight as Python 3.
highlight_language = 'python3'

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.2'

# To perform a Sphinx version check that needs to be more specific than
# major.minor, call `check_sphinx_version("x.y.z")` here.
# check_sphinx_version("1.2.1")

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns.append('_templates')

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog += """
"""

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
#extensions += [
#    'sphinx.ext.mathjax',
#    'sphinx.ext.napoleon',
#    'sphinx.ext.autosummary',
#    'matplotlib.sphinxext.plot_directive',
#]

numpydoc_show_class_members = False
todo_include_todos = False


# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = setup_cfg['name']
author = setup_cfg['author']
copyright = '{0}, {1}'.format(
    datetime.datetime.now().year, setup_cfg['author'])

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.


import_module(setup_cfg['name'])
package = sys.modules[setup_cfg['name']]

# The short X.Y version.
version = package.__version__.split('-', 1)[0]
# The full version, including alpha/beta/rc tags.
release = package.__version__

# The reST default role (used for this markup: `text`) to use for all
# documents.
#default_role = 'any'

# Add this to the top of every file
rst_epilog += '''
.. _classic marx: http://space.mit.edu/cxc/marx/
'''

#autosummary_generate = True

# -- Options for HTML output ----------------------------------------------
# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
#html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# Output file base name for HTML help builder.
htmlhelp_basename = 'Marxsdoc'

html_theme_options = {
    'logotext1': 'MARXS',  # white,  semi-bold
    'logotext2': ':docs',  # orange, light
    'logotext3': ''   # white,  light
}

## -- Options for the edit_on_github extension ----------------------------------------

if eval(setup_cfg.get('edit_on_github')):
    extensions += ['sphinx_astropy.ext.edit_on_github']

    versionmod = __import__(setup_cfg['name'] + '.version')
    edit_on_github_project = setup_cfg['github_project']
    if versionmod.version.release:
        edit_on_github_branch = "v" + versionmod.version.version
    else:
        edit_on_github_branch = "master"

    edit_on_github_source_root = ""
    edit_on_github_doc_root = "docs"

    edit_on_github_skip_regex = '_.*|api/.*'

# -- Resolving issue number to links in changelog -----------------------------
github_issues_url = 'https://github.com/{0}/issues/'.format(setup_cfg['github_project'])



# Licensed under GPL version 3 - see LICENSE.rst
'''

See `ASC FITS File Designers' Guide ASC-FITS-2.1.0 <cxc.harvard.edu/contrib/arots/fits/ascfits.ps>`.
'''
from astropy import time
import astropy.units as u

from ... import __version__
from .data import TLMINMAX, PIXSIZE, NOMINAL_FOCALLENGTH, ODET

def update_header(header, h):
    for elem in h:
        if elem[0] not in header:
            header[elem[0]] = elem[1]

def complete_CC(header, content, hduclass):
    '''Configuration Control Component'''
    h = [
        ('ORIGIN', 'ASC'),
        ("HDUDOC", "ASC-FITS-2.0: Rots, McDowell: ASC FITS File Designers Guide"),
        ("CONTENT", content),
        ("HDUNAME", header['EXTNAME']),
        ]
    for a, b in zip(['HDUCLASS', 'HDUCLAS1', 'HDUCLAS2', 'HDUCLAS3'], hduclass):
        h += [(a, b)]
    update_header(header, h)


def complete_T(header):
    '''Timing component'''
    now = time.Time.now()
    now.format = 'fits'
    nowexp = now + header['EXPOSURE'][0] * u.s
    h = [
        ('DATE', (now.value[:23], 'Date and time of file creation {0}'.format(now.value[23:]))),
        ('DATE-OBS', (now.value[:23], 'TT with clock correction if CLOCKAPP')),
        ('DATE-END', (nowexp.value[:23], 'TT with clock correction if CLOCKAPP')),
        ("TIMESYS", ("TT", "AXAF time will be Terrestrial Time")),
        ("MJDREF", (50814, "MJD of clock start")),
        ("TIMEZERO", (0, "Clock Correction")),
        ("TIMEUNIT", 's'),
        ("BTIMNULL", (0., "Basic Time offset (s)")),
        ("BTIMRATE", (2.5625000912249E-01, "Basic Time clock rate (s / VCDUcount)")),
        ("BTIMDRFT", (2.1806598193841E-17, "Basic Time clock drift (s / VCDUcount^2)")),
        ("BTIMCORR", (0.0000000000000E+00, "Correction applied to Basic Time rate (s)")),
        ("TIMEREF", ("LOCAL", "Time is local for data")),
        ("TASSIGN", ("SATELLITE", "Source of time assignment")),
        ("CLOCKAPP", (True, "Clock correction applied")),
        ("TIERRELA", (1e-9, "Short term clock stability")),
        ("TIERABSO", (1e-4, "Absolute precision of clock correction")),
        ("TIMVERSN", ("ASC-FITS-2.1", "AXAF Fits design document")),
        ("TIMEPIXR", (0., "Time stamp refers to start of bin")),
        ("TIMEDEL", (3.241, "Time resolution of data in seconds")),
        ]
    update_header(header, h)
    # The following keywords depend on the DATE-OBS found in the header, which may differ from
    # the number calculated above, if it was set to a specific date previously.
    tstart = time.Time(header['DATE-OBS'][0], format='fits')
    tstart.format = 'cxcsec'
    header["TSTART"] = (tstart.value, "As in the TIME column: raw space craft clock;")
    header['TSTOP'] = ((tstart + header['EXPOSURE'][0] * u.s).value, "  add TIMEZERO and MJDREF for absolute TT")
    header['OBS-MJD'] = tstart.mjd


def complete_O(header):
    '''Observation info component'''
    h = [("MISSION", ( "AXAF", "Mission is AXAF")),
         ("TELESCOP", ("CHANDRA", "TELESCOPE is Chandra")),
         ("GRATING", ("NONE", "Grating")),
         ('DATACLASS', ('SIMULATED', 'see http://marxs.rtfd.org')),
         ('ONTIME', (header['EXPOSURE'][0], 'Sum of GTIs')),
         ('DTCOR', (1., 'Dead Time Correction')),
         ('LIVETIME', (header['EXPOSURE'][0], 'Ontime multiplied by DTCOR')),
         ('OBSERVER', ('MARXS', 'This is a simulation.')),
         ('FOC_LEN', (NOMINAL_FOCALLENGTH, 'Assumed focal length')),
         ]


DMKEYWORDS = [('MTYPE1', 'chip'), ('MFORM1', 'chipx,chipy'),
              ('MTYPE2', 'tdet'), ('MFORM2', 'tdetx,tdety'),
              ('MTYPE3', 'det'), ('MFORM3', 'detx,dety'),
              ('MTYPE4', 'sky'), ('MFORM4', 'x,y'),
              ('MFORM5', 'RA,DEC'), ('MTYPE5', 'EQPOS')]
'''CIAO data model (DM) keywords that group columns together'''

def add_evt_column_header(header, data):
    '''Add CIAO keywords to header of an eventfile.'''
    # Clean out column related keywords that may not be valid any longer.
    for k in header.iterkeys():
        if k[:5] in ['TCTYP', 'TCRVL', 'TCDLT', 'TCRPX', 'TLMIN', 'TLMAX']:
            del header[k]
    instr = header['INSTRUME'][0]
    if instr not in TLMINMAX.keys():
        raise KeyError('TLMIN and TLMAX not specified for detector {0}'.format(instr))
    colnamesup = [c.upper() for c in data.colnames]
    if len(set(data.colnames)) != len(set(colnamesup)):
        raise KeyError('Fits files are case insensitive. Column names in data must be unique if converted to upper case.')
    tl = TLMINMAX[instr]
    odet = ODET[instr]
    for i, k in enumerate(data.colnames):
        if k.upper() in tl:
            header['TLMIN{0}'.format(i+1)] = tl[k.upper()][0]
            header['TLMAX{0}'.format(i+1)] = tl[k.upper()][1]
    for k in DMKEYWORDS:
        header[k[0]] = k[1]

    # Turn X,Y into a WCS that e.g. ds9 can interpret
    indx = colnamesup.index('X') + 1
    header['TCTYP{0}'.format(indx)] = 'RA---TAN'
    header['TCRVL{0}'.format(indx)] = header['RA_PNT']
    header['TCDLT{0}'.format(indx)] = -PIXSIZE[instr]  # - because RA increases to left
    header['TCRPX{0}'.format(indx)] = odet[0]
    indy = colnamesup.index('Y') + 1
    header['TCTYP{0}'.format(indy)] = 'DEC---TAN'
    header['TCRVL{0}'.format(indy)] = header['DEC_PNT']
    header['TCDLT{0}'.format(indy)] = PIXSIZE[instr]
    header['TCRPX{0}'.format(indy)] = odet[1]
    header['RADECSYS'] = ('ICRS', 'WCS system')


def complete_header(header, data=None, content=['UNKNOWN'], hduclass='UNKNOWN'):
    '''Complete fits header for Chandra fits files.

    This method add the common keywords that are required for all CXC fit files.
    With few exception, the methods will not overwrite existing keywords, but will
    only fill in the value of required keywords if they have not been set before.
    There are certain keywords that cannot be generated without more information
    (e.g. the name of the detector used). These keywords must be set outside of these
    routines.
    The exception where keywords are overwritten is timing information that is contained in
    redundant keywords. This method read "EXPOSURE" and "DATE-OBS" from the header and
    calculates "TSTART", "TSTOP" and "OBS-MJD" from those values.

    Parameters
    ----------
    header : a dictionary-like object
        In most cases, this will be ``photons.meta``.
    data : `astropy.table.Table` or ``None``
        For event tables the header includes keywords that depend on the order of the columns
        in the data (e.g. column 5 and 6 define the WCS). Pass in the full table as data for
        those cases.
    content : string
        Content keyword as specified by ASC
    hduclass : list of stings
        The list can contain 1-4 elements, depending on the data product.
        See appendix A1 in the
        `ASC FITS File Designers' Guide ASC-FITS-2.1.0 <cxc.harvard.edu/contrib/arots/fits/ascfits.ps>`.
    References
    ----------
    `ASC FITS File Designers' Guide ASC-FITS-2.1.0 <cxc.harvard.edu/contrib/arots/fits/ascfits.ps>`.
    '''
    if content.upper().startswith('EVT'):
        add_evt_column_header(header, data)
    complete_CC(header, content=content, hduclass=hduclass)
    complete_T(header)
    complete_O(header)

def sort_columns(photons):
    '''Clean up the order of column names

    Parameters
    ----------
    photons : `astropy.table.Table`
        Event list.

    Returns
    -------
    photons : `astropy.table.Table`
        Event list with columns sorted such that pairs of columns that describe coordinates
        appear together (e.g. "detx" and "dety"). All other columns are sorted alphabetically.
    '''
    colnames = photons.colnames
    # find those pairs that end on x and y like "tdetx, tdety"
    end_x = set([c[:-1] for c in colnames if c[-1] == 'x'])
    end_y = set([c[:-1] for c in colnames if c[-1] == 'y'])
    endxy = end_x.intersection(end_y)
    # columns that are not x/y pairs. Careful: Could be e.g. "energy" without "energx"
    othercol = set(colnames) - set([c+'x' for c in endxy]) - set([c+'y' for c in endxy])
    # Since we are using sets, the order is now random, but CIAO expects certain columns in order
    # When comparing the names, we have to take care of upper case / lower case.
    # That does not matter when we write it to a fits file, but as long as its an astropy Table
    # column names are case sensitive.
    endxy = list(endxy)
    endxy_l = [c.lower() for c in endxy]
    expected_order = ['chip', 'tdet', 'det', '']
    endxy_ordered = []
    for c in expected_order:
        if c in endxy_l:
            index = endxy_l.index(c)
            endxy_ordered.append(endxy.pop(index))
            endxy_l.pop(index)   # Keep endxy_l to have the same order as endxy

    endxy_ordered.extend(endxy)  # Add additional cols not part of the required order.

    ordered_cols = []
    for c in endxy_ordered:
        ordered_cols.extend([c + 'x', c + 'y'])

    # Other pairs we want ordered if present
    pairs = [('ra', 'dec'), ('RA', 'DEC'), ('ra', 'de'), ('RA', 'DE')]
    for p in pairs:
        if p[0] in othercol and p[1] in othercol:
            othercol -= set(p)
            ordered_cols.extend(p)

    # Now put the rest in repeatable order
    othercol = list(othercol)
    othercol.sort()
    ordered_cols.extend(othercol)
    return photons(ordered_cols)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from ..import utils
from transforms3d.affines import compose

def test_random_mat():
    '''Compare multiplication of trans and zoom matrixes with transfomrs3d.

    This test ensures that the same conventions are used for ordering.
    '''
    rot = np.eye(3)
    trans = np.random.rand(3)
    zoom = np.random.rand(3)
    assert np.allclose(np.dot(utils.translation2aff(trans), np.dot(utils.mat2aff(rot), utils.zoom2aff(zoom))),
                       compose(trans, rot, zoom))

def test_random_mat_scalar_zoom():
    '''Compare multiplication of trans and zoom matrixes with transfomrs3d.

    This test ensures that the same conventions are used for ordering.
    '''
    rot = np.eye(3)
    trans = np.random.rand(3)
    zoom = np.random.rand()
    zoommat = np.ones(3) * zoom
    assert np.allclose(np.dot(utils.translation2aff(trans), np.dot(utils.mat2aff(rot), utils.zoom2aff(zoom))),
                       compose(trans, rot, zoommat))

def test_normalize():
    '''Currently only works with 2 d input.'''
    normvec = np.array([[1.,0,0]])
    assert np.all(normvec == utils.norm_vector(normvec))

    vec = np.arange(15, dtype=float).reshape((5,3))
    vecout = np.ones_like(vec)
    for i in range(5):
        vecout[i, :] = vec[i, :] / np.linalg.norm(vec[i, :])
    assert np.allclose(vecout, utils.norm_vector(vec))

def test_anglediff():
    assert np.isclose(utils.anglediff([0, 3.]), 3.)
    assert np.isclose(utils.anglediff([-1, 1]), 2.)
    assert np.isclose(utils.anglediff([1., -1.]), 2 * np.pi - 2.)

def test_angle_between():
    assert utils.angle_between(.1, 0, 2.)
    assert np.all(utils.angle_between(np.array([-.1, 0., 6.1, .2, 6.5]), 5., 1.))
    assert utils.angle_between(2., 6., 1.) == False
    assert utils.angle_between(4., 1., 2.) == False
    assert utils.angle_between(-.5, .5, -3.1) == False
    assert utils.angle_between(0, -np.pi, np.pi) == True



import os
import tempfile

import numpy as np
import pytest
import astropy.units as u
from astropy.table import Table
from astropy.coordinates import SkyCoord
from astropy.utils.data import get_pkg_data_filename

from ..tolerancing import (oneormoreelements,
                           wiggle, moveglobal, moveindividual,
                           varyperiod, varyorderselector, varyattribute,
                           run_tolerances, CaptureResAeff,
                           generate_6d_wigglelist,
                           select_1dof_changed,
                           plot_wiggle, load_and_plot,
                           run_tolerances_for_energies,
                           )
from ...optics import (FlatGrating, OrderSelector, RadialMirrorScatter,
                       RectangleAperture, ThinLens, FlatDetector)

from ...design import RowlandTorus, GratingArrayStructure
from ...utils import generate_test_photons
from ...source import PointSource, FixedPointing
from ...simulator import Sequence

try:
    import matplotlib.pyplot as plt
    HAS_MPL = True
except ImportError:
    HAS_MPL = False


mytorus = RowlandTorus(0.5, 0.5, position=[1.5, 0, -3])

def gsa():
    '''make a parallel structure - fresh for every test'''
    g = GratingArrayStructure(mytorus, d_element=0.1, x_range=[0.5, 1.], radius=[0.1,.2],
                              elem_class=FlatGrating,
                              elem_args={'zoom':0.2, 'd':0.002,
                                         'order_selector': OrderSelector([1])
                                     })
    return g

elempos = np.stack([e.pos4d for e in gsa().elements])

def test_oneormore():
    @oneormoreelements
    def func(a, b, c):
        a.value += 1

    class HoldData():
        def __init__(self, value):
            self.value = value

    obj1 = HoldData(2)
    obj2 = HoldData(4)
    obj3 = HoldData(6)

    listin = [obj2, obj3]

    # First, make sure that func works, otherwise the remaining test is useless.
    func(obj1, 2, c=4)
    assert obj1.value == 3

    func(listin, 'a', None)
    assert listin[0].value == 5
    assert listin[1].value == 7


@pytest.mark.parametrize('function', [wiggle, moveglobal, moveindividual])
def test_change_parallel_elements(function):
    '''Check that parameters work and elements are in fact changed.
    More detailed checks that the type of change is correct are
    implemented as separate tests, but those tests don't check out
    every parameter.
    '''
    g = gsa()
    function(g, 0., 0., 0.)
    assert np.all(np.stack([e.pos4d for e in g.elements]) == elempos)

    for key in ['dx', 'dy', 'dz', 'rx', 'ry', 'rz']:
        d = {key: 1.23}
        function(g, **d)
        assert not np.all(np.stack([e.pos4d for e in g.elements]) == elempos)


def test_moveelements_translate():
    '''Check that the element movers work. If the whole structure is translated
    or individual elements are translated by the same amount, the positions
    should be the same.'''
    g1 = gsa()
    g2 = gsa()
    moveglobal(g1, dy=-20)
    moveindividual(g2, dy=-20)
    assert np.allclose(np.stack([e.pos4d for e in g1.elements]),
                       np.stack([e.pos4d for e in g2.elements]))


def test_moveelements_rotate():
    '''Check that the element movers work.
    Unlike test_moveelements_translate we expect different results because
    there are different center of the rotation.
    This test does not check that the rotation is correct, only that its
    different because the validity of the rotation matrix itself is already
    covered by the tests in the transforms3d package.
    '''
    g1 = gsa()
    g2 = gsa()
    moveglobal(g1, rz=-1, ry=.2)
    moveindividual(g2, rz=-1, ry=.2)
    assert not np.allclose(np.stack([e.pos4d for e in g1.elements]),
                           np.stack([e.pos4d for e in g2.elements]))


def test_wiggle():
    '''Check wiggle function'''
    g = gsa()
    wiggle(g, dx=10, dy=.1)
    diff = elempos - np.stack([e.pos4d for e in g.elements])
    # Given the numbers, wiggle in x must be larger than y
    # This also tests that not all diff number are the same
    # (as they would be with move).
    assert np.std(diff[:, 0, 3]) > np.std(diff[:, 1, 3])


@pytest.mark.parametrize('function', [varyperiod, varyorderselector])
def test_errormessage(function):
    '''Check that check is performed for right type of object.
    Some function just set an attribute and there is no function call after
    that that would fail or do anything if called with the wrong type of object.
    Thus, it's very simple to call these with an object where it does not make
    any sense to apply them. So, they have some error check. Here, we check
    this check.
    '''
    with pytest.raises(ValueError) as e:
        # All functions accept two parameters.
        # Error should be raised before they are used, so the value does not
        # matter
        function(gsa, 1., 2.)

    assert 'does not have' in str(e.value)


def test_gratings_d():
    '''Change the grating constant.'''
    g = gsa()
    varyperiod(g.elements, 1., .1)
    periods = [e._d for e in g.elements]
    assert np.std(periods) > 0.01
    assert np.std(periods) < 5.
    assert np.mean(periods) > .5


def test_scatter():
    '''Check that the right properties are set.'''
    scat = RadialMirrorScatter(inplanescatter=1. * u.arcmin,
                               perpplanescatter=.1 * u.arcmin)
    varyattribute(scat, inplanescatter=2. * u.arcsec,
                  perpplanescatter=.2 * u.degree)
    assert scat.inplanescatter == 2. * u.arcsec
    assert scat.perpplanescatter == .2 * u.degree


def test_errormessage_attribute():
    '''Test error message for generic attributechanger'''
    with pytest.raises(ValueError) as e:
        # All functions accept two parameters.
        # Error should be raised before they are used, so the value does not
        # matter
        varyattribute(gsa, attributenotpresent=1., notpresenteither=2.)

    assert 'does not have' in str(e.value)



def test_orderselector():
    '''Test setting the order selector properties.'''
    photons = generate_test_photons(5)
    grat = FlatGrating(d=1., order_selector=OrderSelector([1]))
    p = grat(photons.copy())
    assert np.all(p['order'] == 1)

    varyorderselector(grat, OrderSelector, [2])
    p = grat(photons.copy())
    assert np.all(p['order'] == 2)


def test_runtolerances():
    '''Test the loop with mock functions.
    This is not a complete functional test, just making sure all calling
    signatures work.
    '''
    photons = generate_test_photons(20)
    grat = FlatGrating(d=1., order_selector=OrderSelector([1]))
    parameters =[{'order_selector': OrderSelector, 'orderlist': [2]},
                 {'order_selector': OrderSelector, 'orderlist': [1, 2], 'p': [.8, 0.]}]

    def afunc(photons):
        return {'meanorder': np.nanmean(photons['order'])}

    out = run_tolerances(photons, grat, varyorderselector, grat,
                         parameters, afunc)

    assert out[0]['meanorder'] == 2
    assert out[1]['meanorder'] == 1
    # check parameters are in output
    assert out[1]['orderlist'] == [1, 2]
    # check original parameter is still intact and can be used again
    # Regression test: If results are inserted into the same dict
    # 'meanorder' will appear which is not valid for varyorderselector
    assert 'meanorder' not in parameters[0]

def test_run_tolerances_for_energies():
    '''For this test, we need to define an instrument. The instrument is not very
    realistic (an X-ray mirror with r=0 won't work), but the point here is just to
    check the tolerancing for several energies. To make that calculation reasonably
    fast, we need to keep the number of elements in the optical system small.
    '''
    coords = SkyCoord(12. * u.deg, -45 * u.deg)
    src = PointSource(coords=coords)
    pnt = FixedPointing(coords=coords)
    aper = RectangleAperture(position=[5000, 0, 0], zoom=[1, 10, 10])
    lens = ThinLens(position=[4900, 0, 0], zoom=[1, 10, 10], focallength=4900)
    grat = FlatGrating(d=.002, order_selector=OrderSelector([0, 1]),
                       position=[4800, 0, 0], zoom=[1, 10, 10])
    det = FlatDetector(zoom=[1, 100, 100])
    instrum = Sequence(elements=[pnt, aper, lens, grat, det])

    parameters = [{'period_mean': 0.003, 'period_sigma': 0.},
                  {'period_mean': 0.004, 'period_sigma': 0.}]

    res = run_tolerances_for_energies(src, [.1, 1] * u.keV,
                                      Sequence(elements=[pnt, aper, lens]),
                                      Sequence(elements=[grat, det]),
                                      varyperiod, grat,
                                      parameters,
                                      CaptureResAeff(orders=[0, 1, 2]),
                                      reset={'period_mean': 0.005,
                                             'period_sigma': 0.},
                                      t_source=1. * u.ks)
    # Check the reset worked
    assert grat._d == 0.005
    # Check both energy have been calculated
    assert 1 in res['energy']
    assert .1 in res['energy']
    assert len(res) == 4
    # check results are reasonable
    assert np.all(res['R'].data[:, 0] == 0)
    assert not np.any(np.isfinite(res['R'].data[:, 2]))
    assert res['R'].data[2, 1] > res['R'].data[0, 1]

def test_capture_res_aeff():
    '''Test the captures res/aeff class.

    Similar to the previous test, this is not a complete functional test,
    but it checks the interfaces.
    The actual function to calculate the effective area is tested elsewhere.
    '''
    p = generate_test_photons(200)
    p['order'] = 0
    p['order'][50:] = 5
    p['xpos'] = -100
    p['xpos'][50:] = np.random.normal(scale=1, size=150)

    resaeff = CaptureResAeff(A_geom=10., order_col='order',
                             orders=[0, 2, 5], dispersion_coord='xpos')
    out = resaeff(p)
    assert np.allclose(out['Aeff'], [2.5, 0., 7.5])
    assert np.isclose(out['Aeff0'], 2.5)
    assert np.isclose(out['Aeffgrat'], 7.5)
    assert len(out['R']) == 3
    assert np.isnan(out['R'][1])

def test_capture_res_aeff_filter():
    '''Ensure that photons with probability 0 will be ignored.
    '''
    p = generate_test_photons(200)
    p['probability'] = 0
    p['order'] = 0
    p['order'][50:] = 5
    p['xpos'] = -100
    p['xpos'][50:] = np.random.normal(scale=1, size=150)

    resaeff = CaptureResAeff(A_geom=10., order_col='order',
                             orders=[0, 2, 5], dispersion_coord='xpos')
    out = resaeff(p)
    assert np.all(out['Aeff'] == 0)
    assert out['Aeff0'] == 0
    assert out['Aeffgrat'] == 0
    assert len(out['R']) == 3
    assert np.isnan(out['R'][2])



def test_6dlist():
    '''Check the list of dicts in 3 translations dof and 3 rotations'''
    cglob, cind = generate_6d_wigglelist([0, 1.] * u.cm, [0., 1.] * u.degree,
                            names=['x', 'y', 'z', 'rx', 'ry', 'rz'])
    assert len(cind) == 7
    assert len(cglob) == 13
    assert set(cind[5].keys()) == set(['x', 'y', 'z', 'rx', 'ry', 'rz'])

    tab = Table(cind)
    for col in ['x', 'y', 'z']:
        assert np.max(tab[col]) == 10
        assert np.min(tab[col]) == 0

    for col in ['x', 'y', 'z']:
        assert np.max(tab[col]) == 10
        assert np.min(tab[col]) == 0

    tab = Table(cglob)
    for col in tab.colnames:
        assert - np.min(tab[col]) == np.max(tab[col])

def test_6d_warning():
    with pytest.warns(UserWarning):
        cglob, cind = generate_6d_wigglelist([1.] * u.cm, [0., 1.] * u.degree)

def test_find_changed():
    '''Test that we find the row where only one parameter was changed.'''
    tab = Table({'par1': [-1, -1, 0, 0, 0, 1],
                 'par2': [-1,  0, 0, 3, 0, 0],
                 'id':   [ 0,  1, 2, 3, 4, 5]})
    t = select_1dof_changed(tab, 'par1', parlist=['par1', 'par2'])
    assert set(t['id']) == set([1, 2, 4, 5])

@pytest.mark.skipif('not HAS_MPL')
def test_plot_wiggle():
    '''Test that wiggle plot works. This does not test that the result
    looks correct, only that running through the plot function does not
    raise any errors.
    This is one of the few plotting functions in the entire package, so
    setting up the infrastructure to compare output pixel-by-pixel does not
    seem worth it as this point.
    '''
    fig, ax = plt.subplots()

    tab = Table({'wave': [1, 1],
                 'dd': [0, 1],
                 'Rgrat': [500, 500],
                 'Aeff': [20, 50]})
    plot_wiggle(tab, 'dd', ['dd'], ax, Aeff_col='Aeff')

@pytest.mark.skipif('not HAS_MPL')
def test_plot_wiggle_exception():
    '''Test that exception is raised for parameters not plotted.
    '''
    fig, ax = plt.subplots()

    tab = Table({'wave': [1, 1],
                 'qwe': [0, 1],
                 'Rgrat': [500, 500],
                 'Aeff': [20, 50]})
    with pytest.raises(ValueError, match='Parameter names should start with'):
        plot_wiggle(tab, 'qwe', ['qwe'], ax, Aeff_col='Aeff')

@pytest.mark.skipif('not HAS_MPL')
def test_plot_6dof():

    tab = Table({'wave': [1, 1, 2, 2, 1, 1, 1, 1],
                 'dd': [0, 1, 0, 2, 0, 0, 0, 0],
                 'rr': [0, 0, 0, 0, 2, 4, 6, 8],
                 'R': np.random.rand(8),
                 'Aeffgrat': np.arange(8)})

    with tempfile.TemporaryDirectory() as tmpdirname:
        name = os.path.join(tmpdirname, 'var_global.fits')
        tab.write(name)
        fig, ax = load_and_plot(name, ['dd', 'rr'], R_col='R')

@pytest.mark.skipif('not HAS_MPL')
def test_plot_6dof_real_file():
    '''Repeat previous test with static data file. This is a more realistic file
    but it takes too long to generate every time. This file is used in the docs
    in design/tolerancing (see docs/pyplot/chandra_tolerancing) so if this test
    breaks, the docs will likely have to be changed, too.
    '''
    filename = get_pkg_data_filename('data/wiggle_global.fits', 'marxs.design.tests')
    fig, ax = load_and_plot(filename)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Column
import astropy.units as u
from astropy.coordinates import SkyCoord

from ..base import SimulationSequenceElement
from ..math.rotations import axangle2mat
from ..math.utils import norm_vector, h2e, e2h


class PointingModel(SimulationSequenceElement):
    '''A base model for all pointing models

    Conventions:

    - All angles (``ra``, ``dec``, and ``roll``) are given in decimal degrees.
    - x-axis points to sky aimpoint.
    - ``roll = 0`` means: z axis points North (measured N -> E).

    For :math:`\delta \pm 90^{\circ}` the :math:`\alpha` value is
    irrelevant for the pointing direction - any right ascension will
    lead to a pointing on the pole. A value for ``ra`` is still
    required, because it determines the orientation of the detector
    plane. Obviously, for pointing straight at the pole, the simple
    interpretation *z axis points north* is meaningless, but the
    combination of ``ra``, ``dec`` and ``roll`` still uniquely
    determines the position of the coordinate system.
    '''
    def add_dir(self, photons):
        linecoords = Column(name='dir', length=len(photons),
                            shape=(4,))
        photons.add_column(linecoords)
        photons['dir'].unit = u.mm
        # Leave everything unset, but chances are I will forget the 4th
        # component. Play safe here.
        photons['dir'][:, 3] = 0

    def process_photons(self, photons):
        self.add_dir(photons)
        return photons



class FixedPointing(PointingModel):
    r'''Transform spacecraft to fixed sky system.

    This matrix transforms from the spacecraft system to a
    right-handed Cartesian system that is defined in the following
    way: the (x,y) plane is defined by the celestial equator, and
    the x-axis points to :math:`(\alpha, \delta) = (0,0)`.

    Parameters
    ----------
    coords : `astropy.coordinates.SkySoord`
        Position of the source on the sky.
    roll : `~astropy.units.quantity.Quantity`
        ``roll = 0`` means: z axis points North (measured N -> E).
    reference_transform : np.array of shape (4, 4)
        By default, photons from an on-axis source come in parallel to the x-axis
        of the coordinate system. Their direction points from x=+inf inwards.
        If the simulation uses a different coordinate system (e.g. the optical
        axis is along the z-axis) set ``reference_transform`` to a matrix that
        performs the conversion.

        The optical axis of the telescope is the normal to the surface of its
        entrance aperture. The pointing needs to know this to determine
        the correct direction of the photons.
        Also, sources that do not shine directly onto the telescope aperture but
        hit it at an angle, will see a smaller projected geometric area.
        This is taken into account by reducing the probability of off-axies photons
        accordingly, and thus this object needs to know the orientation (the
        direction f the optical axis and rotation) of the aperture.

    Notes
    -----
    For :math:`\delta \pm 90^{\circ}` the :math:`\alpha` value is
    irrelevant for the pointing direction - any right ascension will
    lead to a pointing on the pole. A value for ``ra`` is still
    required, because it determines the orientation of the detector
    plane. Obviously, for pointing straight at the pole, the simple
    interpretation *z axis points north* is meaningless, but the
    combination of ``ra``, ``dec`` and ``roll`` still uniquely
    determines the position of the coordinate system.
    '''
    def __init__(self, **kwargs):
        self.coords = kwargs.pop('coords')
        if not self.coords.isscalar:
            raise ValueError("Coordinate must be scalar, not array.")
        self.roll = kwargs.pop('roll', 0. * u.rad)

        self.reference_transform = kwargs.pop('reference_transform', np.eye(4))

        super(FixedPointing, self).__init__(**kwargs)

    @property
    def offset_coos(self):
        '''Return `~astropy.coordinates.SkyOffsetFrame`'''
        return self.coords.skyoffset_frame(rotation=self.roll)

    def photons_dir(self, coos, time):
        '''Calculate direction of photons in homogeneous coordinates.

        Parameters
        ----------
        coos : `astropy.coordiantes.SkyCoord`
            Origin of each photon on the sky
        time : np.array
            Time for each photons in sec

        Returns
        -------
        photons_dir : np.array of shape (n, 4)
            Homogeneous direction vector for each photon
        '''
        photondir = coos.transform_to(self.offset_coos)
        # Minus sign here because photons start at +inf and move towards origin
        photonsdir = norm_vector(-photondir.cartesian.xyz.T)
        return np.einsum('...ij,...j->...i', self.reference_transform, e2h(photonsdir, 0))

    def photons_pol(self, photonsdir, polangle, time):
        '''Calculate a polarization vector for linearly polarized light.

        The current definition cannot handle photons coming exactly from either
        the North pole or the South Pole of the sphere, because the polangle
        definition "North through east" is not well-defined in these positions.

        Parameters
        ----------
        photonsdir : np.array of shape (n, 4)
            Direction of photons
        polangle : np.array
            Polarization angle measured N through E. If polangle has no
            units, it is assumed to be specified in radian.
        time : np.array
            Time for each photons in sec
        '''
        if hasattr(polangle, "unit") and (polangle.unit is not None):
            polangle = polangle.to(u.rad)
        north = SkyCoord(0., 90., unit='deg', frame=self.coords)
        northdir = e2h(north.transform_to(self.offset_coos).cartesian.xyz.T, 0)
        northdir = np.dot(self.reference_transform, northdir)
        n_inskyplane = norm_vector(northdir - photonsdir * np.dot(northdir, photonsdir.T)[:, None])
        e_inskyplane = e2h(np.cross(photonsdir[:, :3], n_inskyplane[:, :3]), 0)
        return  np.cos(polangle)[:, None] * n_inskyplane + np.sin(polangle)[:, None] * e_inskyplane

    def process_photons(self, photons):
        '''
        Parameters
        ----------
        photons : `astropy.table.Table`
        '''
        photons = super(FixedPointing, self).process_photons(photons)
        photons['dir'] = self.photons_dir(SkyCoord(photons['ra'], photons['dec'],
                                                   unit='deg'),
                                          photons['time'].data)
        photons['polarization'] = self.photons_pol(photons['dir'].data,
                                                   photons['polangle'].data,
                                                   photons['time'].data)
        photons.meta['RA_PNT'] = (self.coords.ra.degree, '[deg] Pointing RA')
        photons.meta['DEC_PNT'] = (self.coords.dec.degree, '[deg] Pointing Dec')
        photons.meta['ROLL_PNT'] = (self.roll.to(u.degree).value,
                                    '[deg] Pointing Roll')
        photons.meta['RA_NOM'] = (self.coords.ra.degree, '[deg] Nominal Pointing RA')
        photons.meta['DEC_NOM'] = (self.coords.dec.degree, '[deg] Nominal Pointing Dec')
        photons.meta['ROLL_NOM'] = (self.roll.to(u.degree).value,
                                    '[deg] Nominal Pointing Roll')

        return photons


class JitterPointing(FixedPointing):
    '''Transform spacecraft to fixed sky system.

    This extends `marxs.sourcs.FixedPointing` by adding a random jitter coordinate.
    In this simple implementation the jitter angles applied to two consecutive
    photons are entirely uncorrelated, even if these two photons arrive at the
    same time.
    This class makes the assumption that jitter is small (no change in the projected
    geometric area of the aperture due to jitter).

    Parameters
    ----------
    jitter : `~astropy.units.quantity.Quantity`
        Gaussian sigma of jitter angle
    '''
    def __init__(self, **kwargs):
        self.jitter = np.abs(kwargs.pop('jitter'))
        super(JitterPointing, self).__init__(**kwargs)

    def process_photons(self, photons):
        photons = super(JitterPointing, self).process_photons(photons)
        # Get random jitter direction
        n = len(photons)
        randang = np.random.rand(n) * 2. * np.pi
        ax = np.vstack([np.zeros(n), np.sin(randang), np.cos(randang)]).T
        if self.jitter > 0:
            # For comparison it's often useful to run a model with jitter=0
            # but that would fail np.random.normal(scale=0)
            jitterang = np.random.normal(scale=self.jitter.to(u.radian).value, size=n)
            jitterrot = axangle2mat(ax, jitterang)
            photons['dir'] = e2h(np.einsum('...ij,...i->...j', jitterrot,
                                           h2e(photons['dir'])), 0)
            photons['polarization'] = e2h(np.einsum('...ij,...i->...j', jitterrot,
                                                    h2e(photons['polarization'])), 0)
        return photons



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
from transforms3d.utils import normalized_vector as norm_vec

from ...optics import FlatDetector
from ...simulator import Parallel
from ...math.utils import h2e

from .data import (NOMINAL_FOCALLENGTH, AIMPOINTS, TDET, ODET, PIXSIZE,
                   PIX_CORNER_LSI_PAR, chip2tdet)



ACIS_name = ['I0', 'I1', 'I2', 'I3', 'S0', 'S1', 'S2', 'S3', 'S4', 'S5']
'''names of the 10 ACIS chips'''


class ACISChip(FlatDetector):
    '''A class that defines one chip in the ACIS instrument.'''
    def __init__(self, **kwargs):
        self.TDET = TDET['ACIS']
        self.ODET = ODET['ACIS']
        self.pixsize_in_rad = np.deg2rad(PIXSIZE['ACIS'])
        kwargs['ignore_pixel_warning'] = True
        super(ACISChip, self).__init__(**kwargs)

    @property
    def chip_name(self):
        return 'ACIS-{0}'.format(ACIS_name[self.id_num])

    def specific_process_photons(self, photons, intersect, interpos, intercoos):
        # CHIP and TDET are based on pixel coordinates
        # +1 because Chandra pixel convention is 1 based
        chip = intercoos[intersect, :] / self.pixsize + self.centerpix + 1
        tdet = chip2tdet(chip, self.TDET, self.id_num)
        # DET is based on spacecraft coordiantes, but in observations they need to be derived
        # from pixel coordinates because that's all we have.
        # Here, we already know the spacecraft coordiantes (STF), so we can start from there.
        # I just hope it is consistent and I did not screw up the offsets at some point.
        fc = h2e(interpos[intersect, :])
        mn = fc  # Systems differ only in x-direction
        mn[:, 0] -= NOMINAL_FOCALLENGTH
        x = mn[:, 1] / mn[:, 0] / self.pixsize_in_rad
        y = mn[:, 2] / mn[:, 0] / self.pixsize_in_rad
        detx = self.ODET[0] - x
        dety = self.ODET[1] + y
        theta = np.deg2rad(photons.meta['ROLL_PNT'][0])
        skyx = self.ODET[0] - x * np.cos(theta) + y * np.sin(theta)
        skyy = self.ODET[1] + x * np.sin(theta) + y * np.cos(theta)

        photons.meta['ACSYS1'] = ('CHIP:AXAF-ACIS-1.0', 'reference for chip coord system')
        photons.meta['ACSYS2'] = ('TDET:{0}'.format(self.TDET['version']), 'reference for tiled detector coord system')
        photons.meta['ACSYS3'] = ('DET:ASC-FP-1.1', 'reference for focal plane coord system')
        photons.meta['ACSYS4'] = ('SKY:ASC-FP-1.1', 'reference for sky coord system')
        return {'chipx': chip[:, 0], 'chipy': chip[:, 1],
                'tdetx': tdet[:, 0], 'tdety': tdet[:, 1],
                'detx': detx, 'dety': dety,
                'x': skyx, 'y': skyy,}

class ACIS(Parallel):
    '''The ACIS instrument

    Missing:

    - This currently only implements the ideal **detection**, no PHA, no read-out streaks,
      no pile-up etc.
    - contamination

    '''

    OLSI = np.array([0.684, 0.750, 236.552])
    '''Origin of the LSI in SST coordinates.

    Numbers are taken for the Chandra coordinate memo I at
    http://cxc.harvard.edu/contrib/jcm/ncoords.ps
    '''

    id_col = 'CCD_ID'

    def __init__(self, chips, **kwargs):

        # This stuff is Chandra specific, but applies to HRC, too.
        # Move to a more general class, once the HRC gets implemented.
        self.aimpoint = kwargs.pop('aimpoint')
        self.detoffset = np.array([kwargs.pop(['DetOffsetX'], 0),
                                    0,
                                    kwargs.pop(['DetOffsetY'], 0)])
        # Now the ACIS specific case
        kwargs['elem_pos'] = self.calculate_elempos()
        kwargs['elem_class'] = ACISChip
        # Use 0.024 because that's more consistent with 1024 pix
        kwargs['elem_args'] = {'pixsize': 0.024 } # {'pixsize': 0.023985 }

        super(ACIS, self).__init__(**kwargs)
        self.chips = chips
        self.elements = [self.elements[i] for i in chips]

    def get_corners(self):
        '''Get the coordinates of the ACIS pixel corners.

        Currently, this method reads the datafile included with MARX, but alternatively if could also
        get this information from CALDB.
        It is very unlikely that the coordinates will ever change, so either location will work.

        Returns
        -------
        out : list
            List of dictionaries with entries 'LL', 'LR', 'UR', and 'UL'. The values of the entries are
            3d coordinates of the chip corner in LSI coordinates.
            There is one dictionary per ACIS chip.
        '''
        t = Table.read(PIX_CORNER_LSI_PAR, format='ascii')
        out = []
        for chip in ACIS_name:
            coos = {}
            for corner in ['LL', 'LR', 'UR', 'UL']:
                ind = t['col1'] == 'ACIS-{0}-{1}'.format(chip, corner)
                coos[corner] = np.array([np.float(v) for v in t[ind]['col4'][0][1:-1].split()])
            out.append(coos)
        return out

    def calculate_elempos(self):
        # This stuff is true for HRC, too. Move to more general class, once HRC is implemened.
        corners = self.get_corners()
        pos4d = []
        for i, n in enumerate(ACIS_name):
            # LSI (local science) instrument coordinates
            # notation from coordiante memo: e_x, e_y, e_z is not unit vector!
            p0 = corners[i]['LL']
            e_y = corners[i]['LR'] - p0
            e_z = corners[i]['UL'] - p0
            e_x = np.cross(e_y, e_z)
            center = p0 + 0.5 * e_y + 0.5 * e_z
            # This contains the rotation and the zoom.
            # Note: If I find out that I screwed up the direction of the rotation
            #       I have to treat the zoom separately, because without the .T
            #       it would not work.
            rotlsi = np.vstack([norm_vec(e_x), e_y / 2, e_z / 2]).T
            A = np.eye(4)
            A[:3, :3] = rotlsi
            A[:3, 3] = center
            # Now apply the origin of the LSI system
            A[:3, 3] += self.OLSI
            # see http://cxc.cfa.harvard.edu/proposer/POG/html/chap6.html#tth_chAp6
            A[:3, 3] += self.aimpoint
            # for the coordiante convention - Offset is opposite to direction of coordiante system
            A[:3, 3] -= self.detoffset
            pos4d.append(A)
        return pos4d

    def process_photons(self, photons, *args, **kwargs):
        photons = super(ACIS, self).process_photons(photons, *args, **kwargs)
        photons.meta['SIM_X'] = self.aimpoint[0] - self.detoffset[0]
        photons.meta['SIM_Y'] = self.aimpoint[1] - self.detoffset[1]
        photons.meta['SIM_Z'] = self.aimpoint[2] - self.detoffset[2]
        photons.meta['DETNAM'] = 'ACIS-' + ''.join([str(i) for i in self.chips])
        photons.meta['INSTRUME'] = ('ACIS', 'Instrument')
        return photons



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table
from astropy.io import ascii
from transforms3d import euler
from ...math.utils import norm_vector
from ..multiLayerMirror import MultiLayerMirror, FlatBrewsterMirror

import pytest

def test_photon_reflection():
    ''' tests that three photons are accurately reflected

    TODO: This test should be broken down into smaller components, as should the
    multilayer mirror process photons function.
    '''
    pos = np.tile([1.,0.,0.,1.], (4,1))
    # hitting y = 23mm, 26mm, 24mm; std dev = 0.02, 0.04, 0.01; max = 5.81, 0.42, 6.21; lambda = 3, 6, 4
    dir = np.array([[-1., -1.5, 0., 0],
                    [-1., 1.5, 0., 0],
                    [-1., -0.5, 13., 0],
                    [-1., -1.5, 0., 0]])
    # note: these photons will not hit at 45 degrees (only ok for testing purposes)
    polarization = np.array([[0., 0., 1., 0],
                             [1., 0., 0., 0],
                             [1., 0., 0., 0],
                             [1. / np.sqrt(2.), 0., 1. / np.sqrt(2.), 0.]])
    # z axis is the parallel polarization, v_1, so crossing that with direction (perpendicular to z) is v_2
    polarization[1, 0:3] = np.cross(dir[1,0:3], polarization[0,0:3])
    polarization[1, 0:3] /= np.linalg.norm(polarization[1,0:3])

    photons = Table({'pos': pos,
                     'dir': dir,
                     'energy': [1.23984282 / 3.02, 1.23984282 / 6, 0.4,
                                1.23984282 / 3.02],
                     'polarization': polarization,
                     'probability': [1., 1., 1., 1.]})
    mirror = MultiLayerMirror(reflFile='./marxs/optics/data/testFile_mirror.txt', testedPolarization='./marxs/optics/data/ALSpolarization2.txt', zoom=[1, 24.5, 12])
    photons = mirror(photons)

    # confirm reflection angle
    expected_dir = norm_vector(np.array([[1., -1.5, 0., 0],
                                         [1., 1.5, 0., 0],
                                         [-1., -0.5, 13., 0],
                                         [1., -1.5, 0., 0]]))
    assert np.allclose(norm_vector(np.array(photons['dir'])), expected_dir)

    # confirm reflection probability
    polarizedFile = ascii.read('./marxs/optics/data/ALSpolarization2.txt')
    correctTestPolar = np.interp(1.23984282 / 3.02,
                                 polarizedFile['Photon energy'] / 1000,
                                 polarizedFile['Polarization'])
    expected_prob = np.array([0.0581 * np.exp(-0.5) * 1. / correctTestPolar,
                              0.0042 * 0.,
                              1.,
                              0.5 * 0.0581 * np.exp(-0.5) * 1. / correctTestPolar])
    assert np.allclose(photons['probability'], expected_prob)

    # confirm correct new polarization
    expected_pol = np.array([0., 0, 1., 0])
    simulated_pol = norm_vector(photons['polarization'])
    for i in [0, 3]:
        assert (np.allclose(simulated_pol[i], expected_pol) or
                np.allclose(simulated_pol[i], -expected_pol))
    # prob of relfection is 0, so pol direction is undefined.
    assert np.all(np.isnan(simulated_pol[1][:3]))


def test_photon_reflection2():
    '''more rigorous reflection test'''
    pos = np.array([[0., 0., 1., 1],
                    [0., -0.6, 1., 1],
                    [0., 0., 1., 1],
                    [0., 0.5, 1., 1]])
    # hitting y = 24.5mm, 24mm, 24.5mm, 25mm;
    # ---XXX--- std dev = 0.02, 0.04, 0.01; max = 5.81, 0.42, 6.21; lambda = 3, 6, 4
    dir = np.array([[0., 0., -1., 0],
                    [0., 0.1, -1., 0],
                    [0., 0., -1., 0],
                    [0., 0., -1., 0]])
    # note: these photons will not hit at 45 degrees (only ok for testing purposes)
    polarization = np.array([[0., 1., 0., 0],
                             [0., 1., 0.1, 0],
                             [1., 0., 0., 0],
                             [1., 0., 0., 0]])
    # z axis is the parallel polarization, v_1, so crossing that with direction (perpendicular to z) is v_2

    photons = Table({'pos': pos, 'dir': dir,
                     'energy': [1.23984282 / 3.02, 1.23984282 / 6, 0.4, 0.5],
                     'polarization': polarization,
                     'probability': [1., 1., 1., 1.]})
    a = 2**(-0.5)
    mirror = MultiLayerMirror(reflFile='./marxs/optics/data/testFile_mirror.txt',
                              testedPolarization='./marxs/optics/data/ALSpolarization2.txt',
                              orientation=np.array([[-a, 0, a],[0, -1, 0],[a, 0, a]]))
    photons = mirror(photons)


    # confirm reflection angle
    expected_dir = norm_vector(np.array([[-1., 0., 0., 0], [-1., 0.1, 0., 0], [-1., 0., 0., 0], [-1., 0., 0., 0]]))
    assert np.allclose(norm_vector(np.array(photons['dir'])), expected_dir)

    # This test could be expanded but the numbers in the 'expected' arrays have not been correctly calculated yet.
    #
    # # confirm reflection probability
    # polarizedFile = ascii.read('./marxs/optics/data/ALSpolarization2.txt')
    # correctTestPolar = np.interp(1.23984282 / 3.02, polarizedFile['Photon energy'] / 1000, polarizedFile['Polarization'])
    # expected_prob = np.array([0.0581 * np.exp(-0.5) * 1. / correctTestPolar, 0.0042 * 0., 0.])
    # assert np.allclose(np.array(photons['probability']), expected_prob)
    #
    # # confirm correct new polarization
    # expected_pol = np.array([[0., 1., 0., 0], [1., 1.5, 0., 0], [0., 0., 1., 0], [0., 0., 1., 0]])
    # #expected_pol[1,0:3] = np.cross(photons['dir'][1,0:3] / np.linalg.norm(photons['dir'][1,0:3]), expected_pol[0,0:3])
    # for i in range(0, 2):
    #   assert np.allclose(np.array(photons['polarization'][i]), expected_pol[i]) or np.allclose(np.array(photons['polarization'][i]), -expected_pol[i])

@pytest.mark.parametrize("angle,avgpol", [(0., 0.5), (np.pi/2, 0.)])
def test_double_reflection(angle, avgpol):
    '''Two mirrors

    The first mirror polarizes the light and the second
    one perfectly reflects that polarized light or perfectly absorbs it.
    '''
    pos = np.tile([1., 0., 0., 1.], (8, 1))
    dir = np.tile([-1., 0., 0., 0.], (8, 1))

    ang = np.arange(0, 2. * np.pi, np.pi / 4)
    polarization = np.vstack([np.zeros(8), np.sin(ang), np.cos(ang), np.zeros(8)]).T

    photons = Table({'pos': pos,
                     'dir': dir,
                     'energy': np.ones(8),
                     'polarization': polarization,
                     'probability': np.ones(8)})
    ml1 = FlatBrewsterMirror(orientation=euler.euler2mat(np.pi / 4, 0, 0, 'szxy'))
    ml2 = FlatBrewsterMirror(position=[0, 1, 0],
                             orientation=euler.euler2mat(-np.pi / 4, angle, 0, 'szyx'))
    photons = ml1(photons)
    photons = ml2(photons)
    assert np.isclose(np.mean(photons['probability']), avgpol)



# Licensed under GPL version 3 - see LICENSE.rst
import inspect
from functools import wraps
import collections
from copy import copy
import warnings

import numpy as np
from transforms3d import affines, euler
from astropy.table import Table
from astropy import table
import astropy.units as u
from ..simulator import Parallel
from .uncertainties import generate_facet_uncertainty as genfacun
from ..analysis.gratings import (resolvingpower_from_photonlist,
                                 effectivearea_from_photonlist)
from ..analysis.gratings import AnalysisError

__all__ = ['oneormoreelements',
           'wiggle', 'moveglobal', 'moveindividual',
           'varyperiod', 'varyorderselector', 'varyattribute',
           'run_tolerances', 'run_tolerances_for_energies',
           'CaptureResAeff',
           'generate_6d_wigglelist', 'reset_6d',
           'select_1dof_changed',
           'plot_wiggle', 'load_and_plot',
           ]

def oneormoreelements(func):
    '''Decorator for functions that modify optical elements.

    The functions in this module are written to work on a single optical
    element. This decorator allows them to accept a list of elements or
    a single element.
    '''
    @wraps(func)
    def func_wrapper(elements, *args, **kwargs):
        if isinstance(elements, collections.Iterable):
            for e in elements:
                func(e, *args, **kwargs)
        else:
            func(elements, *args, **kwargs)

    return func_wrapper


@oneormoreelements
def wiggle(e, dx=0, dy=0, dz=0, rx=0., ry=0., rz=0.):
    '''Move and rotate elements around principal axes.

    Parameters
    ----------
    e : `marxs.simulator.Parallel` or list of those elements
        Elements where uncertainties will be set
    dx, dy, dz : float
        accuracy of positioning in x, y, z (in mm) - Gaussian sigma, not FWHM!
    rx, ry, rz : float
        accuracy of positioning. Rotation around x, y, z (in rad) - Gaussian sigma, not FWHM!
    '''
    e.elem_uncertainty = genfacun(len(e.elements), [dx, dy, dz], [rx, ry, rz])
    e.generate_elements()


@oneormoreelements
def moveglobal(e, dx=0, dy=0, dz=0, rx=0., ry=0., rz=0.):
    '''Move and rotate origin of the whole `marxs.simulator.Parallel` object.

    Parameters
    ----------
    e :`marxs.simulator.Parallel` or list of those elements
        Elements where uncertainties will be set
    dx, dy, dz : float
        translation in x, y, z (in mm)
    rx, ry, rz : float
        Rotation around x, y, z (in rad)
    '''
    e.uncertainty = affines.compose([dx, dy, dz],
                                    euler.euler2mat(rx, ry, rz, 'sxyz'),
                                    np.ones(3))
    e.generate_elements()


@oneormoreelements
def moveindividual(e, dx=0, dy=0, dz=0, rx=0, ry=0, rz=0):
    '''Move and rotate all elements of `marxs.simulator.Parallel` object.

    Unlike `~marxs.design.tolerancing.moveglobal` this does not move to
    center of the `~marxs.simulator.Parallel` object, instead it moves
    all elements individually. Unlike `~marxs.design.tolerancing.wiggle`,
    each element is moved by the same amount, not a number drawn from from
    distribution.

    Parameters
    ----------
    e :`marxs.simulator.Parallel` or list of those elements
        Elements where uncertainties will be set
    dx, dy, dz : float
        translation in x, y, z (in mm)
    rx, ry, rz : float
        Rotation around x, y, z (in rad)
    '''
    e.elem_uncertainty = [affines.compose((dx, dy, dz),
                                          euler.euler2mat(rx, ry, rz, 'sxyz'),
                                          np.ones(3))] * len(e.elements)
    e.generate_elements()


@oneormoreelements
def varyattribute(element, **kwargs):
    '''Modify the attributes of an element.

    This function modifies the attributes of an object. The keyword arguments
    are name and value of the attributes to be changed. This can be used for
    a wide variety of MARXS objects where the value of a parameter is stored
    as an instance attribute.

    Parameters
    ----------
    element : object
        Some optical component.
    keywords : it depends

    Examples
    --------
    In a `marxs.optics.RadialMirrorScatter` object, the width of the scattering
    can be modified like this:

    >>> from astropy import units as u
    >>> from marxs.optics import RadialMirrorScatter
    >>> from marxs.design.tolerancing import varyattribute
    >>> rms = RadialMirrorScatter(inplanescatter=2 * u.arcsec, perpplanescatter=0 * u.arcsec)
    >>> varyattribute(rms, inplanescatter=1 * u.arcsec)

    It is usually easy to see which attributes of a
    `marxs.simulator.SimulationSequenceElement` are relevant to be changed in
    tolerancing simulations when looking at the attributes or the implementation
    of those elements.
    '''
    for key, val in kwargs.items():
        # check is needed because key could be misspelled so that we set an attribute
        # that did not exist before and that is never used
        if not hasattr(element, key):
            raise ValueError(f'Object {element} does not have {key} attribute.')
        setattr(element, key, val)


@oneormoreelements
def varyperiod(element, period_mean, period_sigma):
    '''Randomly draw different grating periods for different gratings

    This function needs to be called with gratings as parameters, e.g.

    >>> from marxs.optics import CATGrating
    >>> from marxs.design.tolerancing import varyperiod
    >>> grating = CATGrating(order_selector=None, d=0.001)
    >>> varyperiod(grating, 2e-4, 1e-5)

    and the parameters are expected to have two components:
    center and sigma of a Gaussian distribution for grating contstant d.

    Parameters
    ----------
    element :`marxs.optics.FlatGrating` or similar (or list of those elements)
        Elements where uncertainties will be set
    period_mean : float
        Center of Gaussian (in mm)
    period_sigma : float
        Sigma of Gaussian (in mm)
    '''
    if not hasattr(element, '_d'):
        raise ValueError(f'Object {element} does not have grating period `_d` attribute.')
    element._d = np.random.normal(period_mean, period_sigma)


@oneormoreelements
def varyorderselector(element, order_selector, *args, **kwargs):
    '''Modify the OrderSelector for a grating

    Parameters
    ----------
    element :`marxs.optics.FlatGrating` or similar (or list of those elements)
        Elements where the OrderSelector will be changed
    order_selector : class
        This should be a subclass of `InterpolateRalfTable` which determines how
        the order will be selected. In the case of the default class, the blaze
        angle of an incoming photons will be modified randomly to represent
        small-scale deviations from the flatness of the gratings.
    args, kwargs :
        All other parameters are used to initialize the OrderSelector
    '''
    if not hasattr(element, 'order_selector'):
        raise ValueError(f'Object {element} does not have an order_selector attribute.')
    element.order_selector = order_selector(*args, **kwargs)


def run_tolerances(photons_in, instrum, wigglefunc, wiggleparts,
                   parameters, analyzefunc):
    '''Run tolerancing calculations for a range of parameters

    This function takes an instrument configuration and a function to change
    one aspect of it. For every change it runs a simulations and calculates a
    figure of merit. As the name indicates, this function is designed to derive
    alignment tolerances for certain instrument parts but it might be general
    enough for other parameter studies in instrument design.

    Parameters
    ----------
    photons_in : `astropy.table.Table`
        Input photons list. To speed up the computation, it is useful if photon
        list has been run through all elements of the instrument that are
        located before the first element that is toleranced here.
    instrum : `marxs.simulator.Sequence` object
        An instance of the instrument which contains all elements that
        `photons_in` still have to pass. This can include elements that are
        toleranced in this run and those that are located behind them.
    wigglefunc : callable
        Function that modifies the `instrum` with the following calling signature:
        ``wigglefunc(wiggleparts, pars)`` where ``pars`` is one dict from
        `parameters`. Note that this function is called with `wiggleparts` which
        can be the same as `instrum` or just a subset.
    wiggleparts :  `marxs.base.SimulationSequenceElement` instance
        Element which is modified by `wigglefunc`. Typically this is a subset of
        `instrum`. For example, to tolerance the mirror alignments
        `wiggleparts` would just be the mirror objects, while `instrum` would
        contain all the parts of the instrument that the photons need to run
        though up to the detector.
    parameters : list of dicts
        List of parameter values for calls to `wigglefunc`.
    analyzefunc : callable function or object
        This is called with a photon table and should return a dictionary
        of results. This could be, e.g. a
        `marxs.design.tolerancing.CaptureResAeff` instance.


    Returns
    -------
    result : list of dicts
        Each dict contains parameters and results for one run.

    Notes
    -----
    The format of input and output as lists of dicts is chosen because this
    would work well for a parallel version of this function which could have
    the same interface when it is implemented.
    '''
    out = []
    for i, pars in enumerate(parameters):
        print(f'Working on simulation {i}/{len(parameters)}')
        wigglefunc(wiggleparts, **pars)
        photons = instrum(photons_in.copy())
        cpars = copy(pars)
        cpars.update(analyzefunc(photons))
        out.append(cpars)

    return out


class CaptureResAeff():
    '''Capture resolving power and effective area for a tolerancing simulation.

    Instances of this class can be called with a photon list for a tolerancing
    simulation. The photon list will be analysed for resolving power and
    effective area in a number of relevant orders.

    This is implemented as a class and not a simple function. When the class is
    initialized a number of parameters that are true for any to the
    analysis (e.g. the names of certain columns) are set and saved in the class
    instance.

    The implementation of this class is geared towards instruments with
    gratings but can also serve as an example how a complex analysis that
    derives several different parameters can be implemented.

    Results for effective area and resolving power are reported on a per order
    basis and also summarized for all grating orders combined and the zeroth
    order separately.

    Parameters
    ----------
    A_geom : number
        Geometric area of aperture for the simulations that this instance
        will analyze.
    order_col : string
        Column names for grating orders
    orders : array
        Order numbers to consider in the analysis
    dispersion_coord : string
        Dispersion coordinate for
        `marxs.analysis.gratings.resolvingpower_from_photonlist`

    '''
    def __init__(self, A_geom=1, order_col='order',
                 orders=np.arange(-10, 11), dispersion_coord='det_x'):
        self.A_geom = A_geom
        self.order_col = order_col
        self.orders = np.asanyarray(orders)
        self.dispersion_coord = dispersion_coord

    def __call__(self, photons):
        '''Calculate Aeff and R for an input photon list.

        Parameters
        ----------
        photons : `astropy.table.Table`
            Photon list.

        Returns
        -------
        result : dict
            Dictionary with per-order Aeff and R, as well as values
            summed over all grating orders.
        '''
        aeff = effectivearea_from_photonlist(photons, self.orders, len(photons),
                                             self.A_geom, self.order_col)
        try:
            ind = (np.isfinite(photons[self.dispersion_coord]) &
                   (photons['probability'] > 0))
            res, pos, std = resolvingpower_from_photonlist(photons[ind], self.orders,
                                                           col=self.dispersion_coord,
                                                           zeropos=None,
                                                           ordercol=self.order_col)
        except AnalysisError:
            # Something did not work, e.g. too few photons to find zeroth order
            res = np.nan * np.ones(len(self.orders))

        disporders = self.orders != 0
        # The following lines work for an empty photon list, too.
        aeffgrat = np.sum(aeff[disporders])
        aeff0 = np.sum(aeff[~disporders])

        # Division by 0 causes more nans, so filter those out
        # Also, res is nan if less than 20 photons are detected
        # so we need to filter those out, too.
        ind = disporders & (aeff > 0) & np.isfinite(res)
        if ind.sum() == 0:  # Dispersed spectrum misses detector
            avggratres = np.nan
        else:
            avggratres = np.average(res[ind],
                                    weights=aeff[ind] / aeff[ind].sum())
        return {'Aeff0': aeff0, 'Aeffgrat': aeffgrat, 'Aeff': aeff,
                'Rgrat': avggratres, 'R': res}


def run_tolerances_for_energies(source, energies,
                                instrum_before, instrum_remaining,
                                wigglefunc, wiggleparts,
                                parameters, analyzefunc,
                                reset=None, t_source=1):
    '''Run tolerancing for a grid of parameters and energies

    This function loops over `~marxs.design.tolerancing.run_tolerances` for
    different energies.
    This function takes an instrument configuration and a function to change
    one aspect of it. For every change it runs a simulations and calculates a
    figure of merit. As the name indicates, this function is designed to derive
    alignment tolerances for certain instrument parts but it might be general
    enough for other parameter studies in instrument design.

    There are two nested loops here looping over energy and tolerancing
    parameters. In this case, the outer loop is over energies and then for
    every energy, `~marxs.design.tolerancing.run_tolerances` loops over the
    parameters. This works well where running the ``wigglefunc`` is relatively
    fast, but propagating the photons through ``instrum_before`` is slow and
    minimizes the memory footprint, because only one photon list is in memory
    at any one time. It also implies that runs for the same wiggle parameters
    but different energies are run on different realizations of any random
    draws that are performed by ``wigglefunc``.


    Parameters
    ----------
    source : `marxs.source.Source`
        Source used to generate the photons for every energy. This function
        changes the energy of the source, so the source should be set for
        monoenergetic emission with a constant flux of 1.
    energies : `astropy.units.quantity.Quantity`
        An array of energy values.
    instrum_before : `marxs.simulator.Sequence`
        An instance of the instrument which contains all elements **before**
        the first elements in ``wiggleparts``. The first element should be a
        `marxs.source.pointing.PointingModel`.  In principle,
        ``instrum_before`` can be an empty sequence and the entire instrument
        can be defined in ``instrum_remaining``, however, ``instrum_before`` is
        run only once per energy and thus it can greatly speed up the
        simulation to set this.
    instrum_remaining : `marxs.simulator.Sequence`
        An instance of the instrument which contains all elements not included
        in ``instrum_before``. ``instrum_remaining`` should include elements
        that are toleranced in this run and those that are located behind them.
    wigglefunc, wiggleparts, parameters, analyzefunc :
        These parameters are passed to
        `marxs.design.tolerancing.run_tolerances`. See that function for a
        description of these parameters.
    reset : dict or ``None``
        A dictionary of values for the ``wigglefunc`` function that resets
        the positions or properties of the wiggled elements to their default.
        If ``reset=None``, then the elements affected by ``wigglefunc`` will
        be in the state corresponding to the last entry in ``parameters`` when
        this function exits.
    t_source : int
        parameter for ``source.generate_photons()``. If the source flux is set
        to one, then ``t_source`` determines the number of photons used in the
        tolerancing simulation.

    Returns
    -------
    result : `astropy.table.Table`
        Each row in the table contains energy, wave, parameter values, and
        results from ``analyzefunc`` for a single run.

    '''
    wave = energies.to(u.Angstrom, equivalencies=u.spectral())
    outtabs = []

    for i, e in enumerate(energies):
        source.energy = e
        photons_in = source.generate_photons(t_source)
        photons_in = instrum_before(photons_in)
        data = run_tolerances(photons_in, instrum_remaining,
                              wigglefunc, wiggleparts,
                              parameters, analyzefunc)
        # convert tab into a table.
        # astropy.tables has problems with Quantities as input
        tab = Table([{d: data[i][d].value
                      if isinstance(data[i][d], u.Quantity) else data[i][d]
                      for d in data[i]} for i in range(len(data))])
        tab['energy'] = e
        tab['wave'] = wave[i]
        outtabs.append(tab)
    # Reset so that same instance of instrum can be used again
    if reset is not None:
        wigglefunc(wiggleparts, **reset)

    return table.vstack(outtabs)


def generate_6d_wigglelist(trans, rot,
                           names=['dx', 'dy', 'dz', 'rx', 'ry', 'rz']):
    '''Generate a list of parameters for the wiggle functions in this module.

    This modules contains several wiggle functions such as
    `~marxs.design.tolerances.moveglobal` or
    `~marxs.design.tolerances.wiggle` that expect input for 6 degrees of
    freedom, three translations and three rotations. Commonly, we want to
    study and dof at a time. This function helps with generating lists of
    dicts for that purpose.

    Parameters
    ----------
    trans : `astropy.units.quantity.Quantity`
        Steps for translation. The first element should be 0.
    rot : `astropy.units.quantity.Quantity`
        Steps for rotation. The first element should be 0.

    Returns
    -------
    changeglobal : list of dicts
        This list contains changes in positive and negative directions.
        Use this as input for e.g. `~marxs.design.tolerances.moveglobal`.
    changeindividual : list of dicts
        This list contains only one side, so use this as input for
        e.g. `~marxs.design.tolerances.wiggle` where the actual misalignment
        is drawn from a distribution.

    Examples
    --------
    In this example, we take small steps close to 0 and then increase the
    step size for larger distances, going up to a misalignment of 10 mm in
    linear translation and 2 degrees in rotation::

        >>> import astropy.units as u
        >>> from marxs.design.tolerancing import generate_6d_wigglelist
        >>> trans = [0., .1, .2, .4, .7] * u.cm
        >>> rot = [0., 2., 5., 10., 20] * u.arcmin
        >>> lglob, lind = generate_6d_wigglelist(trans, rot)

    '''
    if (trans.value[0]) != 0 or (rot.value[0] != 0):
        warnings.warn('First element of trans and rot should be 0.')
    n_trans = len(trans)
    n_rot = len(rot)
    n = 3 * n_trans
    changeglobal = np.zeros((n_trans * 2 * 3 + n_rot  * 2 * 3, 6))
    for i in range(3):
        changeglobal[i * n_trans: (i + 1) * n_trans, i] = trans.to(u.mm).value
        changeglobal[n + i * n_rot: n + (i + 1) * n_rot, i + 3] = rot.to(u.rad).value

    half = changeglobal.shape[0] / 2
    changeglobal[int(half):, :] = - changeglobal[:int(half), :]
    changeindividual = changeglobal[: int(half), :]
    # Remove multiple entries with [0,0,0,0, ...]
    changeglobal = np.unique(changeglobal, axis=0)
    changeindividual = np.unique(changeindividual, axis=0)

    # numpy array to list of dicts
    changeglobal = [dict(zip(names, row)) for row in changeglobal]
    changeindividual = [dict(zip(names, row)) for row in changeindividual]

    return changeglobal, changeindividual


reset_6d = {'dx': 0., 'dy': 0., 'dz': 0., 'rx': 0., 'ry': 0., 'rz': 0.}
'''The neutral element for a 6d wigglelist. Can be used to reset uncertainties.'''


def select_1dof_changed(table, par,
                        parlist=['dx', 'dy', 'dz', 'rx', 'ry', 'rz']):
    '''Select subset of tolerancing table with changes in 1 dof only

    This function selects a subset of a table where all parameters other
    than the parameter selected right now are zero.

    This is a helper function for analyzing and plotting results.

    Parameters
    ----------
    table : `astropy.table.Table`
        Table with wiggle results
    par : string
        Name of parameter to be selected
    parlist : list of strings
        Name of all parameters in ``table``

    Returns
    -------
    tab : `astropy.table.Table`
        Filtered table
    '''
    pars = set(parlist)
    ind = np.ones(len(table), dtype=bool)
    for p in pars - set([par]):
        ind *= table[p] == 0
    return table[ind]


def plot_wiggle(tab, par, parlist, ax, axt=None,
                R_col='Rgrat', Aeff_col='Aeffgrat',
                axes_facecolor='w'):
    '''Plotting function for overview plot wiggeling 1 dof at the time.

    For parameters starting with "d" (e.g. "dx", "dy", "dz"), the plot axes
    will be labeled as a shift, for parameters tarting with "r" as rotation.

    Parameters
    ----------
    table : `astropy.table.Table`
        Table with wiggle results
    par : string
        Name of parameter to be plotted
    parlist : list of strings
        Name of all parameters in ``table``
    ax : `matplotlib.axes.Axes`
        Axis object to plot into.
    axt : ``None`` or  `matplotlib.axes.Axes`
        If this is ``None``, twin axis are created to show resolving power
        and effective area in one plot. Alternatively, a second axes instance
        can be given here.
    R_col : string
        Column name in ``tab`` that hold the resolving power to be plotted.
        Default is set to work with `marxs.design.tolerancing.CaptureResAeff`.
    Aeff_col : string
        Column name in ``tab`` that hold the effective area to be plotted.
        Default is set to work with `marxs.design.tolerancing.CaptureResAeff`.
    axes_facecolor : any matplotlib color specification
        Color for the background in the plot.
    '''
    import matplotlib.pyplot as plt

    t = select_1dof_changed(tab, par, parlist)
    t.sort(par)
    t_wave = t.group_by('wave')
    if axt is None:
        axt = ax.twinx()

    for key, g in zip(t_wave.groups.keys, t_wave.groups):
        if par[0] == 'd':
            x = g[par]
        elif par[0] == 'r':
            x = np.rad2deg(g[par].data)
        else:
            raise ValueError("Don't know how to plot {}. Parameter names should start with 'd' for shifts and 'r' for rotations.".format(par))

        ax.plot(x, g[R_col], label='{:3.1f} $\AA$'.format(key[0]), lw=1.5)
        axt.plot(x, g[Aeff_col], ':', label='{:2.0f} $\AA$'.format(key[0]), lw=2)
    ax.set_ylabel('Resolving power (solid lines)')
    axt.set_ylabel('$A_{eff}$ [cm$^2$] (dotted lines)')
    if par[0] == 'd':
        ax.set_xlabel('shift [mm]')
        ax.set_title('Shift along {}'.format(par[1]))
    elif par[0] == 'r':
        ax.set_xlabel('Rotation [degree]')
        ax.set_title('Rotation around {}'.format(par[1]))

    for a in [ax, axt]:
        a.set_facecolor(axes_facecolor)
        a.set_axisbelow(True)
        a.grid(axis='x', c='1.0', lw=2, ls='solid')


wiggle_plot_facecolors = {'global': '0.9',
                          'individual': (1.0, 0.9, 0.9)}
'''Default background colors for wiggle overview plots.

If the key of the dict matches part of the filename, the color listed in
the dict is applied.
'''


def load_and_plot(filename, parlist=['dx', 'dy', 'dz', 'rx', 'ry', 'rz'], **kwargs):
    '''Load a table with wiggle results and make default plot

    This is a function to generate a quicklook image with many
    hardcoded defaults for figure size, colors etc.
    In particular, this function is written for the display of
    6d plots which vary 6 degrees of freedom, one at a time.

    The color for the background in the plot is set depending on the filename
    using the ``string : color`` assignments in
    `~marxs.design.tolerancing.wiggle_plot_facecolors`. No fancy regexp based
    match is applied, this is simply a check with ``in``.

    Parameters
    ----------
    filename : string
        Path to a file with data that can be plotted by
        `~marxs.design.tolerancing.plot_wiggle`.

    parlist : list of strings
        Name of all parameters in ``table``.
        This function only plots six of them.

    Returns
    -------
    tab : `astropy.table.Table`
        Table of data read from ``filename``
    fig : `matplotlib.figure.Figure`
        Figure with plot.
    kwargs :
        All other parameters are passed to
        `~marxs.design.tolerancing.plot_wiggle`.
    '''
    import matplotlib.pyplot as plt

    tab = Table.read(filename)

    if 'axis_facecolor' not in kwargs:
        for n, c in wiggle_plot_facecolors.items():
            if n in filename:
                kwargs['axes_facecolor'] = c

    fig = plt.figure(figsize=(12, 8))
    fig.subplots_adjust(wspace=.6, hspace=.3)
    for i, par in enumerate(parlist):
        ax = fig.add_subplot(2, 3, i + 1)
        plot_wiggle(tab, par, parlist, ax, **kwargs)

    return tab, fig



# Licensed under GPL version 3 - see LICENSE.rst
'''Plot routines for display with `three.js <http://threejs.org>`__.

Each routine adds strings to a file and thus each routing requires a writable file
object as an argument. This file can then be included in a webpage that loads
the three.js javascript package.

Note that routine to display a torus is modified
relative to the official three.js release to allow more parameters; the modified
version is included in the MARXS source code.
Also, note that the output file is not a valid webside by itself. Instead, it is a list of
javascript commands, that needs to be included into a website after the nesessary setup
and package loading.
'''
import numpy as np
from astropy.utils.decorators import format_doc

from . import utils

listofproperties = {'Material': ['id', 'name', 'opacity', 'transparent', 'blending', 'blendSrc',
                                 'blendDst', 'BlendEquation', 'depthTest', 'depthWrite',
                                 'polygonOffset', 'plygonOffsetFactor', 'polygonOffsetUnits',
                                 'alphaTest', 'clippingPlanes', 'clipShadows', 'overdraw',
                                 'visible', 'side', 'needsUpdate'],
                    'MeshStandardMaterial': ['color', 'roughness', 'metalness', 'map', 'lightMap',
                                             'lightMapIntensity', 'aoMap', 'aoMapIntensity',
                                             'emissive', 'emissiveMap', 'emissiveIntensity',
                                             'bumpMap', 'bumpMapScale', 'normalMap', 'normalMapScale',
                                             'displacementMap', 'displacementScale',
                                             'displacementBias', 'roughnessMap', 'metalnessMap',
                                             'alphaMap', 'envMap', 'envMapIntensity', 'refractionRatio',
                                             'fog', 'shading', 'wireframe', 'wireframeLinewidth',
                                             'wireframeLinecap', 'wireframeLinejoin', 'vertexColors',
                                             'skinning', 'morphTargets', 'morphNormals'],
                    'LineBasicMaterial' : ['color', 'linewidth', 'linecap', 'linejoin', 'vertexColors',
                                           'fog'],
                    }


def array2string(array):
    '''Flatten array and return string representation for insertion into js.'''
    return np.array2string(array.flatten(), max_line_width=1000, separator=',',
                           formatter={'float_kind': '{0:.1f}'.format})

def materialdict(display, material):
    '''Construct a string that can be pasted into a javascript template to describe a three.js material.

    Parameters
    ----------
    display : dict
        Dictionary of properties. All properties that have a same name as the property of
        the relevant material in javascript are included, all others are ignored.
    material : string
        Name of the material in three.js (Only materials used currently in marxs are supported).

    Returns
    -------
    spec : dict
        material specification
    '''
    spec = {}
    for k in display:
        if (k in listofproperties['Material']) or (k in listofproperties[material]):
            # now special cases that need to be transformed in some way
            if k == 'color':
                spec['color'] = utils.color_tuple_to_hex(display[k]).replace('0x', '#')
            else:
                spec[k] = display[k]
    if ('opacity' in display) and not ('transparent' in display):
        spec['transparent'] = 'true'
    return spec


def materialspec(display, material):
    '''Construct a string that can be pasted into a javascript template to describe a three.js material.

    Parameters
    ----------
    display : dict
        Dictionary of properties. All properties that have a same name as the property of
        the relevant material in javascript are included, all others are ignored.
    material : string
        Name of the material in three.js (Only materials used currently in marxs are supported).

    Returns
    -------
    spec : string
        String that can be pasted into javasript files.
    '''
    matdict = materialdict(display, material)
    spec = ['{0} : {1}'.format(k, matdict[v]) for k in matdict]
    return ', '.join(spec)

def _format_plot_rays_input(data, scalar, cmap, prop):
    # The number of points per line
    N = data.shape[1]
    # number of lines
    n = data.shape[0]

    if scalar is None:
        s = np.zeros((n,  N))
    elif scalar.shape == (n, ):
        s = np.tile(scalar, (N, 1)).T
    elif scalar.shape == (n, N):
        s = scalar
    else:
        raise ValueError('Scalar quantity for each point must have shape ({0},) or ({0}, {1})'.format(n, N))

    import matplotlib.pyplot as plt
    cmap = plt.get_cmap(cmap)
    normalizer = plt.Normalize()
    s_rgb = cmap(normalizer(s))

    if 'vertexColors' not in prop:
        prop['vertexColors'] = 2 # 2 = 'THREE.VertexColors'

    return data, s_rgb, prop, n

def plot_rays(data, outfile, scalar=None, cmap=None,
              prop={}):
    '''Plot lines for simulated rays.

    Parameters
    ----------
    data : np.array of shape(n, N, 3)
        where n is the number of rays, N the number of positions per ray and
        the last dimension is the (x,y,z) of an Eukledian position vector.
    scalar : None or nd.array of shape (n,) or (n, N)
        This quantity is used to color the rays. If ``None`` all rays will have the same
        color. If it has n elements, each ray will have exactly one color (e.g. color
        according to the energy of the ray), if it has n*N elements, rays will be
        multicolored.
    outfile : file object
        Output javascript code is written to this file.
    prop : dict
        keyword arguments for line material.
    cmap : `matplotlib.colors.Colormap` instance or string or None
        `matplotlib` color maps are used to convert ``scalar`` values to rgb colors.
        If ``None`` the default matplotlib colormap will be used, otherwise the colormap
        can be specified in this keyword.
    '''
    data, s_rgb, prop, n = _format_plot_rays_input(data, scalar, cmap, prob)
    material = materialspec(prop, 'LineBasicMaterial')

    for i in range(n):
        positions = array2string(data[i, :, :])
        colors = array2string(s_rgb[i, :, :3])
        outfile.write('''
        var geometry = new THREE.BufferGeometry();
        var material = new THREE.LineBasicMaterial({{ {material} }});
        var positions = new Float32Array({positions});
        var colors = new Float32Array({colors});
        geometry.addAttribute( 'position', new THREE.BufferAttribute( positions, 3 ) );
        geometry.addAttribute( 'color', new THREE.BufferAttribute( colors, 3 ) );
        geometry.computeBoundingSphere();
        mesh = new THREE.Line( geometry, material );
        scene.add( mesh );'''.format(positions=positions, colors=colors, material=material))


doc_plot='''
    {__doc__}

    Parameters
    ----------
    obj : `marxs.base.MarxsElement`
        The element that should be plotted.
    display : dict of None
        Dictionary with display settings.
    outfile : file handle
        Output is written to this file.
'''

@format_doc(doc_plot)
def plot_object(obj, display, outfile, **kwargs):
    '''Plot any MARXS object with three.js as backend.

    This method will inspect the object that is passed in and select the
    correct plotting method for its shape. Javascript code to generate the correct
    represenation in three.js is added to the open file ``outfile``.
    '''
    if 'outfile' not in kwargs:
        raise TypeError('Required argument "outfile" missing.')
    out = plot_object_general(plot_registry, obj, display=display,outfile=outfile, **kwargs)
    return out


@format_doc(doc_plot)
def container(obj, display, outfile):
    '''Recursivey output three.js commands to print all elements of a container.

    Output of each element can be a dict (if it is a leaf) or a list
    (if it is a container). We need to flatten the list here to avoid
    arbitrarily deep recursion.
    '''
    return [plot_obj(e, e.display, outfile) for e in obj.elements]

@format_doc(doc_plot)
def box(obs, display, outfile):
    '''Output three.js commands to include a box-shaped optical element.'''
    matrixstring = ', '.join([str(i) for i in obj.pos4d.flatten()])
    if not ('side' in display):
        display['side'] = 'THREE.DoubleSide'
    materialspec = materialspec(display, 'MeshStandardMaterial')
    outfile.write('''
    var geometry = new THREE.BoxGeometry( 2, 2, 2 );
    var material = new THREE.MeshStandardMaterial( {{ {materialspec} }} );
    var mesh = new THREE.Mesh( geometry, material );
    mesh.matrixAutoUpdate = false;
    mesh.matrix.set({matrix});
    scene.add( mesh );'''.format(materialspec=materialspec,
                                 matrix=matrixstring))

@format_doc(doc_plot)
def triangulation(obj, display, outfile):
    '''Output commands for a plane with an inner hole.'''
    xyz, triangles = obj.geometry.triangulate(display)
    materialspec = materialspec(display, 'MeshStandardMaterial')
    outfile.write('// {}\n'.format(obj.name))
    outfile.write('var geometry = new THREE.BufferGeometry(); \n')
    outfile.write('var vertices = new Float32Array([')
    for row in xyz:
        outfile.write('{0}, {1}, {2},'.format(row[0], row[1], row[2]))
    outfile.write(''']);
    // itemSize = 3 because there are 3 values (components) per vertex
    geometry.addAttribute( 'position', new THREE.BufferAttribute( vertices, 3 ) );
    ''')
    outfile.write('var faces = new Uint16Array([')
    for row in triangles:
        outfile.write('{0}, {1}, {2}, '.format(row[0], row[1], row[2]))
    outfile.write(''']);
    // itemSize = 3 because there are 3 values (components) per triangle
    geometry.setIndex(new THREE.BufferAttribute( faces, 1 ) );
    ''')

    outfile.write('''var material = new THREE.MeshStandardMaterial({{ {materialspec} }});
        var mesh = new THREE.Mesh( geometry, material );
        scene.add( mesh );
        '''.format(materialspec=materialspec))

@format_doc(doc_plot)
def torus(obj, display, outfile):
    '''Output commands to display (part of) a torus.

    The plot range for theta and phi is taken from the values for ``coo1`` and ``coo2``
    in the ``display`` dictionary. There entries should be a list of value if less then
    the full torus (ranging from 0 to 2 pi in each coordinate) is desired for the plot.
    '''
    theta = display.get('coo1', [0, 2 * np.pi])
    phi = display.get('coo2', [0, 2 * np.pi])
    materialspec = materialspec(obj.display, 'MeshStandardMaterial')
    torusparameters = '{0}, {1}, {2}, {3}, {4}, {5}, {6}, {7}'.format(obj.R, obj.r,
                                                                      int(np.rad2deg(theta[1])),
                                                                      int(np.rad2deg(phi[1])),
                                                                      theta[1],
                                                                      theta[0],
                                                                      phi[1],
                                                                      phi[0])
    rot = np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1.]])
    matrixstring = ', '.join([str(i) for i in np.dot(obj.pos4d, rot).flatten()])

    outfile.write('''
    var geometry = new THREE.ModifiedTorusBufferGeometry({torusparameters});
    var material = new THREE.MeshStandardMaterial({{ {materialspec} }});
    var mesh = new THREE.Mesh( geometry, material );
    mesh.matrixAutoUpdate = false;
    mesh.matrix.set({matrix});
    scene.add( mesh );'''.format(materialspec=materialspec,
                                 torusparameters=torusparameters,
                                 matrix=matrixstring))

plot_registry = {'triangulation': triangulation,
                 'torus': torus,
                 'box': box,
                 'container': container,
                 }



# Licensed under GPL version 3 - see LICENSE.rst
from ..base import DocMeta

def test_docsting_inheritance():
    class A(metaclass=DocMeta):
        '''class doc here'''

        A = 5 # int: attribute A

        def func(a, b):
            '''Function that does stuff with a and b.'''
            pass

    class B(A):
        def func(a,b,c):
            pass

    assert 'Function that does stuff' in B.func.__doc__



import os
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u
from .... import optics, source
from ..hrma_py import Aperture, HRMA
from .. import HETG

parfile = os.path.join(os.path.dirname(optics.__file__), 'hrma.par')
marx = optics.MarxMirror(parfile)
aperpy = Aperture()
mirrpy = HRMA()


def test_FWHM():
    '''Compare properties of PSF to MARX C code'''
    coords = SkyCoord(25., 45., unit=u.deg)
    src = source.PointSource(coords=coords)
    pointing = source.FixedPointing(coords=coords)
    p = src(10 * u.ks)
    p = pointing(p)
    pmarx = marx(p.copy())
    ppy = aperpy(p.copy())
    ppy = mirrpy(ppy)

    fp = optics.FlatDetector()
    pmarx = fp(pmarx)
    ppy = fp(ppy)

    assert np.isclose(np.nanstd(pmarx['det_x']), np.nanstd(ppy['det_x']), rtol=0.25)
    assert np.isclose(np.nanstd(pmarx['det_y']), np.nanstd(ppy['det_y']), rtol=0.25)
    assert np.isclose(np.nanmean(ppy['det_x']), 0, atol=1e-3)


def test_most_photons_hit_grating():
    '''If the radii of the mirror shells are correct, most photons should hit a
    facet in the HETG.'''
    coords = SkyCoord(25., 45., unit=u.deg)
    src = source.PointSource(coords=coords)
    pointing = source.FixedPointing(coords=coords)
    p = src(10. * u.ks)
    p = pointing(p)
    p = aperpy(p.copy())
    p = mirrpy(p)
    hetg = HETG()
    p = hetg(p)
    assert p['probability'][p['facet'] > 0].sum() > 5000.



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure.

from astropy.version import version as astropy_version

# As of Astropy 3.0, the pytest plugins provided by Astropy are
# automatically made available when Astropy is installed. This means it's
# not necessary to import them here, but we still need to import global
# variables that are used for configuration.
from astropy.tests.plugins.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS

from astropy.tests.helper import enable_deprecations_as_exceptions

## Uncomment the following line to treat all DeprecationWarnings as
## exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
## as follow (although default should work for most cases).
## To ignore some packages that produce deprecation warnings on import
## (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
## 'setuptools'), add:
##     modules_to_ignore_on_import=['module_1', 'module_2']
## To ignore some specific deprecation warning messages for Python version
## MAJOR.MINOR or later, add:
##     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# enable_deprecations_as_exceptions()

## Uncomment and customize the following lines to add/remove entries from
## the list of packages for which version numbers are displayed when running
## the tests. Making it pass for KeyError is essential in some cases when
## the package uses other astropy affiliated packages.
try:
    PYTEST_HEADER_MODULES['Astropy'] = 'astropy'
    PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'
    del PYTEST_HEADER_MODULES['h5py']
    del PYTEST_HEADER_MODULES['Pandas']
except KeyError:
    pass

# Uncomment the following lines to display the version number of the
# package rather than the version number of Astropy in the top line when
# running the tests.
import os

# This is to figure out the package version, rather than
# using Astropy's
try:
    from .version import version
except ImportError:
    version = 'dev'

try:
    packagename = os.path.basename(os.path.dirname(__file__))
    TESTED_VERSIONS[packagename] = version
except NameError:   # Needed to support Astropy <= 1.0.0
    pass

import matplotlib
matplotlib.use("Agg")



import numpy as np
import matplotlib.pyplot as plt
from astropy.table import QTable
import astropy.units as u
from marxs.source import Source
en = np.arange(0.5, 7., .5) * u.keV
fluxperbin = en.value**(-2) / u.s / u.cm**2 / u.keV

# Input as astropy QTable
tablespectrum = QTable([en, fluxperbin], names=['energy', 'fluxdensity'])
s = Source(energy=tablespectrum, name='table')


fig = plt.figure()

photons = s.generate_photons(1e3 * u.s)
plt.hist(photons['energy'], histtype='step', label=s.name, bins=20)

leg = plt.legend()
lab = plt.xlabel('Energy [keV]')
lab = plt.ylabel('Counts / bin')



import numpy as np

from ...utils import generate_test_photons
from ...optics import ThinLens, FlatGrating, FlatDetector, OrderSelector
from ..gratings import (resolvingpower_from_photonlist,
                        effectivearea_from_photonlist,
                        )

def test_resolvingpower_from_photonlist():
    '''Regression test.

    [#159] left behind an undefined ``filterfunc``. This is fixed and a
    regression test added [#165].
    '''
    p = generate_test_photons(1000)
    for i in [1, 2]:
        p['pos'][:, i] = -0.1 + 0.2 * np.random.random(len(p))
    lens = ThinLens(focallength=1)
    p = lens(p)
    orders = np.array([-1, 0, 1])
    grat = FlatGrating(position=[-0.01, 0, 0], d=1e-4,
                       order_selector=OrderSelector(orders))
    p = grat(p)
    det = FlatDetector(position=[-1, 0, 0])
    p = det(p)
    res0, pos0, std0 = resolvingpower_from_photonlist(p, orders, col='det_x')
    assert res0[0] > 10
    assert res0[1] < 1
    assert res0[2] > 10
    res, pos, std = resolvingpower_from_photonlist(p, orders, col='det_x', zeropos=0)
    assert res[0] > 10
    assert res[1] < 1
    assert res[2] > 10
    # y is the cross-dispersion direction...
    resy, posy, stdy = resolvingpower_from_photonlist(p, orders, col='det_y', zeropos=0)
    assert np.all(resy < 1.)
    assert np.allclose(posy, 0, atol=1e-4)


def test_aeff_from_photonlist():
    '''Check the effective area calculation'''
    p = generate_test_photons(5)
    p['a'] = [0, 1, 2, 2, 1]
    p['probability'] = [.1, 1., 1., .3, .5]
    # orders in random order, just in case
    aeff = effectivearea_from_photonlist(p, [2, 1, 0, -1], 5,
                                         A_geom=2., ordercol='a')
    assert aeff[3] == 0
    assert np.isclose(aeff[2], .04)
    assert np.isclose(aeff[1], .6)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from transforms3d.utils import normalized_vector

__all__ = ['ex2vec_fix', 'axangle2mat']

def ex2vec_fix(e1, efix):
    '''Rotate x-axis to e1, use efix to break rotation ambiguity.

    This function calcalates the rotation matrix that rotates the x-axis to
    ``e1``, i.e. it rotates the normal of the y,z-plane to a new plane where
    ``e1`` is the normal. This still leaves the rotation angle of the plane
    free. This function breaks the degeneracy by keeping a vector that is
    coplanar with the normal and fix coplanar with fix and the new normal.

    The purpose of this function is best explained by an example:
    Imagine a plane in the y, z plane. We want to rotate the plane, such that ``e1``
    will be a new normal of the plane. However, since the orientation of the plane
    is not fixed, there is an infinite number of ways to do this.
    This ambiguity can be broken if we additionally require that the rotated
    :math:`\hat e_y` will be coplanar with ``fix``.

    Inspired by the algorithm described in:
    https://www.fastgraph.com/makegames/3drotation/

    Parameters
    ----------
    e1 : np.array of shape (3, )
        new normal of plane
    efix : np.array of shape (3, )
        Vector to break rotational ambiguity.

    Returns
    -------
    rot : np.array of shape (3, 3)
        Rotation matrix
    '''
    e1 = normalized_vector(e1)
    efix = normalized_vector(efix)
    if np.allclose(e1, efix) or np.allclose(e1, -efix):
        raise ValueError('Input vectors are parallel - Rotation matrix is ambiguous.')
    rot = np.empty((3, 3))
    rot[:, 0] = e1
    rot[:, 1] = normalized_vector(efix - np.dot(efix, e1) * e1)
    rot[:, 2] = np.cross(rot[:, 0], rot[:, 1])
    return rot


def axangle2mat(axes, angles, is_normalized=False):
    ''' Rotation matrix for rotation angle `angle` around `axis`

    This is a vectorized version of the routine of the same name in
    ``transforms3d``.

    Parameters
    ----------
    axes : np.array of shape (N, 3)
       vector specifying axis for rotation.
    angle : np.array
       angle of rotation in radians.
    is_normalized : bool, optional
       True if `axis` is already normalized (has norm of 1).  Default False.

    Returns
    -------
    mat : array shape (N, 3,3)
       rotation matrices for specified rotation

    Notes
    -----
    From: http://en.wikipedia.org/wiki/Rotation_matrix#Axis_and_angle
    '''
    if len(angles) != axes.shape[0]:
        raise ValueError('There must be one angle for each axes vector.')

    if not is_normalized:
        axes = axes / np.linalg.norm(axes, axis=1)[:, None]
    c = np.cos(angles); s = np.sin(angles); C = 1-c
    x = axes[:, 0]; y = axes[:, 1]; z = axes[:, 2]
    xs = x*s;   ys = y*s;   zs = z*s
    xC = x*C;   yC = y*C;   zC = z*C
    xyC = x*yC; yzC = y*zC; zxC = z*xC
    return np.array([
            [ x*xC+c,   xyC-zs,   zxC+ys ],
            [ xyC+zs,   y*yC+c,   yzC-xs ],
            [ zxC-ys,   yzC+xs,   z*zC+c ]]).swapaxes(0,2).swapaxes(1,2)



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
import astropy.units as u
from scipy.stats import normaltest
import pytest

from ..scatter import RadialMirrorScatter, RandomGaussianScatter
from ..detector import FlatDetector
from ...utils import generate_test_photons

def test_distribution_of_scattered_rays():
    '''Check that scattered rays have a normal distribution.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    rms = RadialMirrorScatter(inplanescatter=0.1 * u.rad)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    assert np.all(p['det_y'] == 0)
    stat, pval = normaltest(p['det_x'])
    assert pval > 0.5
    assert np.allclose(np.std(p['det_x']), np.arctan(0.1), rtol=0.1)

def test_distribution_scatered_rays_turned():
    '''Test that the in-plane scattering plane is correctly calculated.

    The photons are positioned 90 deg from the example above.
    Also, do some perpendicular to the plan scattering, that's missing
    from the test above.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 0., 1., 1.]), (500, 1))
    rms = RadialMirrorScatter(inplanescatter=0.1 * u.rad,
                              perpplanescatter=0.01 * u.rad)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    # This is in-plane-scatter
    assert np.allclose(np.std(p['det_y']), np.arctan(0.1), rtol=0.1)
    # This is scatter perpendicular to the plane.
    assert np.allclose(np.std(p['det_x']), np.arctan(0.01), rtol=0.1)

def test_scatter_0_01():
    '''Test that scatter works if one or both scatter directions are set
    to 0. These cases are special cased, so that need an extra test.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    rms = RadialMirrorScatter(inplanescatter=0 * u.rad,
                              perpplanescatter=0.1 * u.rad)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    assert np.all(p['det_x'] == 1.0)
    assert np.allclose(np.std(p['det_y']), np.arctan(0.1), rtol=0.1)

def test_scatter_0_0():
    '''Test that scatter works if one or both scatter directions are set
    to 0. These cases are special cased, so that need an extra test.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    rms = RadialMirrorScatter(inplanescatter=0 * u.rad,
                              perpplanescatter=0 * u.rad)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    assert np.all(p['det_x'] == 1.0)
    assert np.all(p['det_y'] == 0)


def test_gaussscatter_0():
    '''Test that scatter works if scatter sigma set
    to 0. This case is special cased, so that needs an extra test.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    rms = RandomGaussianScatter(scatter=0 * u.rad)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)

    assert np.all(p['det_x'] == 1.0)
    assert np.all(p['det_y'] == 0)


def test_gaussiandistribution_of_scattered_rays():
    '''Check that scattered rays have a normal distribution.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    rms = RandomGaussianScatter(scatter=1e-5 * u.rad)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)
    # Calculate distance from undisturbed position
    d = np.sqrt((p['det_x'] - np.mean(p['det_x']))**2 +
                (p['det_y'] - np.mean(p['det_y']))**2)
    d = d * np.sign(p['det_x'] - np.mean(p['det_x']))

    stat, pval = normaltest(d)
    assert pval > 0.02
    assert np.isclose(np.std(d), 1e-5, rtol=0.1)


def test_scatteredfunction():
    '''Check that scattered rays have a normal distribution.'''
    photons = generate_test_photons(500)
    photons['pos'] = np.tile(np.array([0., 1., 0., 1.]), (500, 1))
    photons['time'] = np.arange(len(photons), dtype=float) / 500.

    def timescatter(photons, intersect, interpos, intercoos):
        return photons['time'] * u.degree

    rms = RandomGaussianScatter(scatter=timescatter)
    det = FlatDetector(position=[-1, 0, 0], zoom=1000)

    p = rms(photons)
    p = det(p)
    # Calculate distance from undisturbed position
    d = np.sqrt((p['det_x'] - np.mean(p['det_x']))**2 +
                (p['det_y'] - np.mean(p['det_y']))**2)
    d = d * np.sign(p['det_x'] - np.mean(p['det_x']))

    assert np.std(d[:200]) < 2 * np.std(d[:300])


def test_scatterarg_missing():
    '''Scatter argument is required, unless overriden by a derived class.'''
    with pytest.raises(ValueError) as e:
        rms = RandomGaussianScatter()

    assert 'Keyword "scatter" missing.' in str(e.value)


def test_derivedScatterClass():
    '''Check that overriding default scatter works.'''
    class Derived(RandomGaussianScatter):
        scatter = 1 * u.arcsec

    # This should work without scatter keyword
    rms = Derived()

    # This should raise a warning, but set the value
    with pytest.warns(UserWarning, match="Overriding class level"):
        rms = Derived(scatter=5 * u.arcsec)
    assert rms.scatter == 5 * u.arcsec



# Licensed under GPL version 3 - see LICENSE.rst
'''Tests for calling marx the right way.

Test in this module are not meant to check that the marx c code works - that really
is the responsibility of the marx maintainer in a separate repository.
Instead, it does a few spot check to ensure that the c modules are called in the right
way. In particular, it holds regression tests for issues that were already discovered
and fixed.
'''
import numpy as np
from scipy.stats import ks_2samp
from astropy.coordinates import SkyCoord
import astropy.units as u

import marxs
import marxs.source
import marxs.optics.marx

def test_noexplicettimedependence():
    '''In an older implementation, the first half of all photons went of
    mirror shell 0, which turned out to be due to ``sorted_index`` being a int,
    while marx expects an *unsigned* int.
    '''
    mysource = marxs.source.PointSource(coords=SkyCoord(30., 30., unit="deg"))
    photons = mysource.generate_photons(1 * u.ks)
    mypointing = marxs.source.FixedPointing(coords=SkyCoord(30, 30., unit='deg'))
    photons = mypointing(photons)

    marxm = marxs.optics.marx.MarxMirror('./marxs/optics/hrma.par', position=np.array([0., 0,0]))
    photons = marxm(photons)
    ks, p_value = ks_2samp(photons['mirror_shell'][:400], photons['mirror_shell'][600:])
    assert p_value > 1e-5
